from nltk.tag import EspressoTagger

if __name__ == '__main__':
    sent = "나는 배가 고프다. 나는 아름다운 강산에 살고있다." 
    tagger = EspressoTagger()
    print()
    print(tagger.tag('pos', sent))
    print("dependency :")
    print(tagger.tag('dependency', sent))
    print('ner :')
    ner = tagger.tag('ner', sent)
    print(ner)
    print()
    print()
    print('wsd :')
    print(tagger.tag('wsd', sent))
    print()
    #print('srl :')
    #print(tagger.tag('srl', sent))



from nltk.alignment import NeedlemanWunsch, SmithWaterman, Hirschberg, LongestCommonSubsequence, LongestCommonSubstring, DTW
from nltk.distance import LevenshteinEditDistance, HammingDistance, DamerauLevenshteinDistance, WassersteinDistance
from nltk.similarity import CosineSimilarity, LCSubstringSimilarity, LCSubsequenceSimilarity, JaroSimilarity
from nltk.tokenize import sent_tokenize, word_tokenize, syllable_tokenize
from nltk.search import NaiveSearch, RabinKarpSearch, KMPSearch, BoyerMooreSearch, FaissSearch
from nltk.metrics import BERTScore, BARTScore, DefaultMetric
from nltk import pos_tag, nouns, word_segmentor, pos_tag_with_verb_form
import numpy as np
from typing import List
import torch

def demo():
    str1 = '기존에 제품이 장기간 사용으로 손상'
    str2 = '장기간 사용으로 제품이 손상'

    # result1, result2 = NeedlemanWunsch().get_alignment(str1, str2)
    # print(result1, '\n', result2)

    result1, result2 = SmithWaterman().get_alignment(str1, str2)
    print(f"{result1}\n{result2}")

    # result1, result2 = Hirschberg().get_alignment(str1, str2)
    # print(f"{result1}\n{result2}")

    # result = DTW().get_alignment_path(str1, str2)
    # print(result)

    # result = LongestCommonSubsequence().compute(str1, str2)
    # print(result)

    # result = LongestCommonSubstring().compute(str1, str2)
    # print("-------LongestCommonSubstring-------")
    # print(result)
    # print("------------------------------------")
    # print()

def demo2():
    str1 = '나는 학생이다.'
    str2 = '그는 선생님이다.'

    result = BARTScore().compute([str1], [str2])
    print("-------BARTScore-------")
    print(result)
    print("-----------------------")
    print()

def demo3():
    str1 = '나는 학생이다.'
    str2 = '그는 선생님이다.'
    model_name = 'bert-base-uncased'
    result = BERTScore(model_name_or_path=model_name, lang='kor', num_layers=12).compute([str1], [str2])

    print("model name: ", model_name)
    print("-------BERTScore-------")
    print(result)
    print("-----------------------")
    print()

def demo4():
    demo_setences = ['제가 나와 있는 곳은 경남 거제시 옥포동 덕포 해수욕장에 나와 있습니다.']
    for sen in demo_setences:
        print(word_tokenize(sen, "korean"))
        print(pos_tag(sen, lang='kor'))

def demo5():
    str1 = '나는 학생이다.'
    str2 = '그는 선생님이다.'

    # result = LevenshteinEditDistance().compute(str1, str2)

    # result = HammingDistance().compute(str1, str2)


    result = DamerauLevenshteinDistance().compute(str1, str2)

    print("-------DamerauLevenshteinDistance-------")
    print(result)
    print("----------------------------------------")
    print()

def demo6():
    x1 = np.array([1, 2, 3, 4, 5])
    x2 = np.array([3, 7, 8, 3, 1])

    result = CosineSimilarity().compute(x1, x2)

    print("-------CosineSimilarity-------")
    print(result)
    print("------------------------------")
    print()

def demo7():
    str1 = '나는 학생이다.'
    str2 = '그는 선생님이다.'

    result = LCSubstringSimilarity().compute(str1, str2)

    print("-------LCSubstringSimilarity-------")
    print(result)
    print("-----------------------------------")
    print()

    result = LCSubsequenceSimilarity().compute(str1, str2)

    print("-------LCSubsequenceSimilarity-------")
    print(result)
    print("--------------------------------------")
    print()

    result = JaroSimilarity().compute(str1, str2)

    print("-------JaroSimilarity-------")
    print(result)
    print("----------------------------")
    print()


def demo8():
    pattern = "학생"
    str1 = '나는 학생이다.'

    result = NaiveSearch().search(pattern, str1)
    print(result)

    result = RabinKarpSearch().search(pattern, str1)
    print(result)

    result = KMPSearch().search(pattern, str1)
    print(result)

    result = BoyerMooreSearch().search(pattern, str1)
    print(result)

def demo9():
    faiss = FaissSearch(model_name_or_path = 'skt/kobert-base-v1', tokenizer_name_or_path = 'skt/kobert-base-v1')
    corpus = {
        'text': [
                "오늘은 날씨가 매우 덥습니다.",
                "저는 음악을 듣는 것을 좋아합니다.",
                "한국 음식 중에서 떡볶이가 제일 맛있습니다.",
                "도서관에서 책을 읽는 건 좋은 취미입니다.",
                "내일은 친구와 영화를 보러 갈 거예요.",
                "여름 휴가 때 해변에 가서 수영하고 싶어요.",
                "한국의 문화는 다양하고 흥미로워요.",
                "피아노 연주는 나를 편안하게 해줍니다.",
                "공원에서 산책하면 스트레스가 풀립니다.",
                "요즘 드라마를 많이 시청하고 있어요.",
                "커피가 일상에서 필수입니다.",
                "새로운 언어를 배우는 것은 어려운 일이에요.",
                "가을에 단풍 구경을 가고 싶어요.",
                "요리를 만들면 집안이 좋아보입니다.",
                "휴대폰 없이 하루를 보내는 것이 쉽지 않아요.",
                "스포츠를 하면 건강에 좋습니다.",
                "고양이와 개 중에 어떤 동물을 좋아하세요?"
                "천천히 걸어가면서 풍경을 감상하는 것이 좋아요.",
                "일주일에 한 번은 가족과 모임을 가요.",
                "공부할 때 집중력을 높이는 방법이 있을까요?",
                "봄에 꽃들이 피어날 때가 기대되요.",
                "여행 가방을 챙기고 싶어서 설레여요.",
                "사진 찍는 걸 좋아하는데, 카메라가 필요해요.",
                "다음 주에 시험이 있어서 공부해야 해요.",
                "운동을 하면 몸이 가벼워집니다.",
                "좋은 책을 읽으면 마음이 풍요로워져요.",
                "새로운 음악을 발견하면 기분이 좋아져요.",
                "미술 전시회에 가면 예술을 감상할 수 있어요.",
                "친구들과 함께 시간을 보내는 건 즐거워요.",
                "자전거 타면 바람을 맞으면서 즐거워집니다."
        ],
    }
    print(faiss.initialize_corpus(corpus=corpus, section='text', embedding_type='mean_pooling', save_path='/Users/dowon/Test/test.json'))
    query = "오늘은 날씨가 매우 춥다."
    top_k = 5
    result = faiss.search(query, top_k)
    print(result)

def faiss_test():
    faiss = FaissSearch(model_name_or_path = 'klue/bert-base')
    result = TextReader("/Users/dowon/Test/sentence1.txt").read()
    id = 0
    
    for i in result:
        print(i)
        i = i.replace('\n', '')
        print(i)
        i = "i am test"
        print(faiss.get_embeddings(text=i, num_workers=10).detach().cpu().numpy())
        id += 1
        if id ==3:
            break

def faiss_save_test():
    faiss = FaissSearch(model_name_or_path = '/Users/dowon/test_model/trained_model/', tokenizer_name_or_path = '/Users/dowon/test_model/trained_model/')
    faiss.load_dataset_from_json('/Users/dowon/Test/test.json')
    faiss.embedding_type = 'mean_pooling'
    # faiss.load_faiss_index(index_name='embeddings',file_path='/Users/dowon/Test/test_index.json')
    faiss.add_faiss_index(column_name='embeddings')
    query = "오늘은 날시가 매우 춥다."
    top_k = 5
    result = faiss.search(query, top_k)
    print(result)
    

def demo10():
    metric = DefaultMetric()
    y_true = [1, 3, 3, 5, 5,1]
    y_pred = [1, 2, 3, 4, 5,2]
    str1 = "i am teacher"
    str2 = "he is student"
    print(metric.precision_score(y_true, y_pred, "macro"))

def demo11():
    print("\nBegin Wasserstein distance demo ")

    P =  np.array([0.6, 0.1, 0.1, 0.1, 0.1])
    Q1 = np.array([0.1, 0.1, 0.6, 0.1, 0.1])
    Q2 = np.array([0.1, 0.1, 0.1, 0.1, 0.6])

    P = torch.from_numpy(P)
    Q1 = torch.from_numpy(Q1)
    Q2 = torch.from_numpy(Q2)
    kl_p_q1 = WassersteinDistance().compute_kullback(P, Q1)
    kl_p_q2 = WassersteinDistance().compute_kullback(P, Q2)

    wass_p_q1 = WassersteinDistance().compute_wasserstein(P, Q1)
    wass_p_q2 = WassersteinDistance().compute_wasserstein(P, Q2)

    jesson_p_q1 = WassersteinDistance().compute_jesson_shannon(P, Q1)
    jesson_p_q2 = WassersteinDistance().compute_jesson_shannon(P, Q2)


    print("\nKullback-Leibler distances: ")
    print("P to Q1 : %0.4f " % kl_p_q1)
    print("P to Q2 : %0.4f " % kl_p_q2)

    print("\nWasserstein distances: ")
    print("P to Q1 : %0.4f " % wass_p_q1)
    print("P to Q2 : %0.4f " % wass_p_q2)

    print("\nJesson-Shannon distances: ")
    print("P to Q1 : %0.4f " % jesson_p_q1)
    print("P to Q2 : %0.4f " % jesson_p_q2)
    
    print("\nEnd demo ")

def demo12():
	y_pred = [5, 2, 4, 1, 3, 2, 5, 6, 7]
	y_true = [1, 3, 6, 7, 1, 5]

	user = [[5, 3, 2], [9, 1, 2], [3, 5, 6], [7, 2, 1]]
	h_pred = [[15, 6, 21, 3], [15, 77, 23, 14], [51, 23, 21, 2], [53, 2, 1, 5]]

	metric = DefaultMetric()
	print(metric.precision_at_k(y_true,  y_pred, 3))
	print(metric.recall_at_k(y_true,y_pred, 3))
	print(metric.hit_rate_at_k(user, h_pred, 1))



class TextReader:
    def __init__(self, path: str):
        self.path = path

    def read(self) -> List[str]:
        with open(self.path, 'r') as f:
            return f.readlines()


if __name__=="__main__":
    # demo()
    # demo2()
    # demo3()
    #demo4()
    # demo5()
    # demo6()
    # demo7()
    # demo8()
    # demo9()
    # faiss_test()
    # faiss_save_test()
    # demo10()
    demo11()
    #demo12()



# -*- coding: utf-8 -*-
# Natural Language Toolkit: Tokenizers
#
# Copyright (C) 2001-2020 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
#         Steven Bird <stevenbird1@gmail.com> (minor additions)
# Contributors: matthewmc, clouds56
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT

r"""
NLTK Tokenizer Package

Tokenizers divide strings into lists of substrings.  For example,
tokenizers can be used to find the words and punctuation in a string:

    >>> from nltk.tokenize import word_tokenize
    >>> s = '''Good muffins cost $3.88\nin New York.  Please buy me
    ... two of them.\n\nThanks.'''
    >>> word_tokenize(s)
    ['Good', 'muffins', 'cost', '$', '3.88', 'in', 'New', 'York', '.',
    'Please', 'buy', 'me', 'two', 'of', 'them', '.', 'Thanks', '.']

This particular tokenizer requires the Punkt sentence tokenization
models to be installed. NLTK also provides a simpler,
regular-expression based tokenizer, which splits text on whitespace
and punctuation:

    >>> from nltk.tokenize import wordpunct_tokenize
    >>> wordpunct_tokenize(s)
    ['Good', 'muffins', 'cost', '$', '3', '.', '88', 'in', 'New', 'York', '.',
    'Please', 'buy', 'me', 'two', 'of', 'them', '.', 'Thanks', '.']

We can also operate at the level of sentences, using the sentence
tokenizer directly as follows:

    >>> from nltk.tokenize import sent_tokenize, word_tokenize
    >>> sent_tokenize(s)
    ['Good muffins cost $3.88\nin New York.', 'Please buy me\ntwo of them.', 'Thanks.']
    >>> [word_tokenize(t) for t in sent_tokenize(s)]
    [['Good', 'muffins', 'cost', '$', '3.88', 'in', 'New', 'York', '.'],
    ['Please', 'buy', 'me', 'two', 'of', 'them', '.'], ['Thanks', '.']]

Caution: when tokenizing a Unicode string, make sure you are not
using an encoded version of the string (it may be necessary to
decode it first, e.g. with ``s.decode("utf8")``.

NLTK tokenizers can produce token-spans, represented as tuples of integers
having the same semantics as string slices, to support efficient comparison
of tokenizers.  (These methods are implemented as generators.)

    >>> from nltk.tokenize import WhitespaceTokenizer
    >>> list(WhitespaceTokenizer().span_tokenize(s))
    [(0, 4), (5, 12), (13, 17), (18, 23), (24, 26), (27, 30), (31, 36), (38, 44),
    (45, 48), (49, 51), (52, 55), (56, 58), (59, 64), (66, 73)]

There are numerous ways to tokenize text.  If you need more control over
tokenization, see the other methods provided in this package.

For further information, please see Chapter 3 of the NLTK book.
"""
from nltkor.tokenize.ko_tokenize import Ko_tokenize



"""
string2string similarity
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""


"""
    This class contains the original implementation of the BERTScore algorithm by Zhang et al. (2020).

    BERTScore: Evaluating Text Generation with BERT

    @inproceedings{bertscore2020,
        title={BERTScore: Evaluating Text Generation with BERT},
        author={Tianyi Zhang* and Varsha Kishore* and Felix Wu* and Kilian Q. Weinberger and Yoav Artzi},
        booktitle={International Conference on Learning Representations},
        year={2020},
        url={https://openreview.net/forum?id=SkeHuCVFDr}
    }

    Disclaimer:
        This code is adapted from https://github.com/Tiiiger/bert_score
"""

from typing import List, Union, Optional, Tuple

import os
import sys
import time
from collections import defaultdict

try:
    import pandas as pd
    import torch
    from bert_score.utils import (bert_cos_score_idf, get_hash, 
                              get_idf_dict, get_model, get_tokenizer,
                              lang2model, model2layers)
    from nltk.search.kobert_tokenizer import KoBERTTokenizer
except ImportError:
    raise Exception("""You need to install Library 
    please pip install below Libaries
    \t pip install torch
    \t pip install pandas
    \t pip install bert_score
    """)

# import torch
# from bert_score.utils import (bert_cos_score_idf, get_hash, 
#                               get_idf_dict, get_model, get_tokenizer,
#                               lang2model, model2layers)
# from nltk.search.kobert_tokenizer import KoBERTTokenizer





class BERTScore:
    """
    This class implements the BERTScore algorithm.
    """
    
    def __init__(self, 
        model_name_or_path: str = None,
        lang: str = None,
        num_layers: int = None,
        all_layers: bool = False,
        use_fast_tokenizer: bool = False,
        device: str = 'cpu',
        baseline_path: str = None,
    ) -> None:
        r"""
        This function initializes the BERTScore class, which computes the BERTScore between two texts.

        Arguments:
            model_name_or_path (str): BERT model type to use (e.g., bert-base-uncased).
            lang (str): Language of the texts (e.g., en).
            num_layers (int): Number of layers to use.
            all_layers (bool): Whether to use all layers
            use_fast_tokenizer (bool): Whether to use the fast tokenizer.
            device (str): Device to use (e.g., cpu or cuda).
            baseline_path (str): Path to the baseline file.

        Returns:
            None

        Raises:
            ValueError: If model_name_or_path and lang are both None.

        .. attention::

            If you use this class, please make sure to cite the following paper:
        
            .. code-block:: latex

                @inproceedings{bertscore2020,
                    title={BERTScore: Evaluating Text Generation with BERT},
                    author={Tianyi Zhang* and Varsha Kishore* and Felix Wu* and Kilian Q. Weinberger and Yoav Artzi},
                    booktitle={International Conference on Learning Representations},
                    year={2020},
                    url={https://openreview.net/forum?id=SkeHuCVFDr}
                }

        
        .. note::
            * If model_name_or_path is not specified, use the default model for the language.
            * If num_layers is not specified, use the default number of layers.
            * If device is not specified, use the GPU if available, otherwise use the CPU.
            * If baseline_path is not specified, use the default baseline file.
        """

        # Check the arguments
        if model_name_or_path is None and lang is None:
            raise ValueError("You must specify either model_name_or_path or lang")
        
        # Set the attributes
        self.model_name_or_path = model_name_or_path
        self.lang = lang
        self.num_layers = num_layers
        self.all_layers = all_layers
        self.use_fast_tokenizer = use_fast_tokenizer
        self.baseline_path = baseline_path

        # If model_name_or_path is not specified, use the default model for the language
        if self.model_name_or_path is None:
            self.lang = lang.lower()
            self.model_name_or_path = lang2model[self.lang]

        # If num_layers is not specified, use the default number of layers
        if num_layers is None:
            self.num_layers = model2layers[self.model_name_or_path]
        
        # Set the device
        self.device = device
        if self.device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # Load model and tokenizer
        if self.model_name_or_path == 'skt/kobert-base-v1':
            self.tokenizer = KoBERTTokenizer.from_pretrained(self.model_name_or_path)
        else:
            self.tokenizer = get_tokenizer(self.model_name_or_path, self.use_fast_tokenizer)
        self.model = get_model(self.model_name_or_path, self.num_layers, self.all_layers)
        self.model.eval()
        self.model.to(device)



    # Compute the BERTScore between source sentences and target sentences
    def compute(self,
        source_sentences: List[str],
        target_sentences: Union[List[str], List[List[str]]],
        batch_size: int = 4,
        idf: bool = False,
        nthreads: int = 4,
        return_hash: bool = False,
        rescale_with_baseline: bool = False,
        verbose: bool = False,
    ) -> Union[dict, Optional[str]]:
        """
        This function scores the source sentences based on their similarity to the target sentences using BERTScore.

        Arguments:
            source_sentences (list of str): candidate sentences
            target_sentences (list of str or list of list of str): reference sentences
            batch_size (int): bert score processing batch size
            idf (bool or dict): use idf weighting, can also be a precomputed idf_dict
            nthreads (int): number of threads
            return_hash (bool): return hashcode of the setting
            rescale_with_baseline (bool): rescale bertscore with pre-computed baseline
            verbose (bool): turn on intermediate status update

        Returns:
            (Dict[str, Tensor], Optional[str]): A dictionary containing the precision, recall, and F1 score, and the hashcode (if return_hash is True).
                where the precision, recall, and F1 score are tensors of shape (len(source_sentences),

        Raises:
            ValueError: If the number of source sentences and target sentences do not match.
        """

        # Check the arguments
        if len(source_sentences) != len(target_sentences):
            raise ValueError("The number of candidates and references do not match")
        
        # If the target sentences are grouped, flatten them
        ref_group_boundaries = None
        if not isinstance(target_sentences[0], str):
            ref_group_boundaries = []
            ori_source_sentences, ori_target_sentences = source_sentences, target_sentences
            source_sentences, target_sentences = [], []
            count = 0
            for cand, ref_group in zip(ori_source_sentences, ori_target_sentences):
                source_sentences += [cand] * len(ref_group)
                target_sentences += ref_group
                ref_group_boundaries.append((count, count + len(ref_group)))
                count += len(ref_group)

        if rescale_with_baseline and self.baseline_path is None:
            raise ValueError("Need to specify baseline_path when rescaling with baseline")

        # Get the IDF dict
        if not idf:
            idf_dict = defaultdict(lambda: 1.0)
            # set idf for [SEP] and [CLS] to 0
            idf_dict[self.tokenizer.sep_token_id] = 0
            idf_dict[self.tokenizer.cls_token_id] = 0
        elif isinstance(idf, dict):
            if verbose:
                print("using predefined IDF dict...")
            idf_dict = idf
        else:
            if verbose:
                print("preparing IDF dict...")
            start = time.perf_counter()
            idf_dict = get_idf_dict(target_sentences, self.tokenizer, nthreads=nthreads)
            if verbose:
                print("done in {:.2f} seconds".format(time.perf_counter() - start))

        if verbose:
            print("calculating scores...")
        
        start = time.perf_counter()

        # Get all the predictions
        all_preds = bert_cos_score_idf(
            model = self.model,
            refs = target_sentences,
            hyps = source_sentences,
            tokenizer= self.tokenizer,
            idf_dict = idf_dict,
            verbose = verbose,
            device = self.device,
            batch_size=batch_size,
            all_layers=self.all_layers,
        ).cpu()

        # If the target sentences are grouped, take the max score
        if ref_group_boundaries is not None:
            max_preds = []
            for beg, end in ref_group_boundaries:
                max_preds.append(all_preds[beg:end].max(dim=0)[0])
            all_preds = torch.stack(max_preds, dim=0)

        # Rescale with baseline
        use_custom_baseline = self.baseline_path is not None
        if rescale_with_baseline:
            if self.baseline_path is None:
                self.baseline_path = os.path.join(
                    os.path.dirname(__file__), f"rescale_baseline/{self.lang}/{self.model_name_or_path}.tsv"
                )
            if os.path.isfile(self.baseline_path):
                if not self.all_layers:
                    baselines = torch.from_numpy(
                        pd.read_csv(self.baseline_path).iloc[self.num_layers].to_numpy()
                    )[1:].float()
                else:
                    baselines = (
                        torch.from_numpy(pd.read_csv(self.baseline_path).to_numpy())[:, 1:]
                        .unsqueeze(1)
                        .float()
                    )

                all_preds = (all_preds - baselines) / (1 - baselines)
            else:
                print(
                    f"Warning: Baseline not Found for {self.model_name_or_path} on {self.lang} at {self.baseline_path}",
                    file=sys.stderr,
                )

        # Get the final output
        out = all_preds[..., 0], all_preds[..., 1], all_preds[..., 2]  # P, R, F
        scores = {
            "precision": out[0].numpy(),
            "recall": out[1].numpy(),
            "f1": out[2].numpy(),
        }
        
        # Print the time
        if verbose:
            time_diff = time.perf_counter() - start
            print(
                f"done in {time_diff:.2f} seconds, {len(target_sentences) / time_diff:.2f} sentences/sec"
            )

        # If return hash, return both the output and the hash
        if return_hash:
            return tuple(
                [
                    scores,
                    get_hash(
                        self.model_name_or_path,
                        self.num_layers,
                        idf,
                        rescale_with_baseline,
                        use_custom_baseline=use_custom_baseline,
                        use_fast_tokenizer=self.use_fast_tokenizer,
                    ),
                ]
            )
        # Otherwise, just return the output
        return scores




def demo():
    demo_setences = [
        ("I am a student", "He is a teacher"),
        ("나는 학생이다", "그는 선생님이다"),
        ("점심에 온기동에서 삼겹차슈덮밥을 먹었다.", "저녁에 피나치공에서 피자와 치킨을 먹었다.")
    ]
    for str1, str2 in demo_setences:
        print("demo : ", BERTScore(model_name_or_path='bert-base-uncased', lang='en', num_layers=12).compute([str1], [str2]))

if __name__ == "__main__":
    demo()


# -*- coding: utf-8 -*-

import itertools
from collections import Counter, OrderedDict as OD
from nltkor import Kor_char

class WordDictionary(dict):
	"""
	Class to store words and their corresponding indices in
	the network lookup table. Also deals with padding and
	maps rare words to a special index.
	"""

	padding_left = '*LEFT*'
	padding_right = '*RIGHT*'
	rare = '*RARE*'

	number_transformation = {ord(c): '9' for c in '012345678'}
	#english_transformation = {ord(c): 'a' for c in 'abcdefghijklmnopqrstuvwxyz'}
	# 한자 처리 필요 korChar.hanja_syllable() 사용 , utils.py 참조

	def __init__(self, tokens, size=None, minimum_occurrences=None, wordlist=None):
		"""
		Fills a dictionary (to be used for indexing) with the most
		common words in the given text.

		:param tokens: Either a list of tokens or a list of lists of tokens
			(each token represented as a string).
		:param size: Maximum number of token indices
			(not including paddings, rare, etc.).
		:param minimum_occurrences: The minimum number of occurrences a token must
			have in order to be included.
		:param wordlist: Use this list of words to build the dictionary. Overrides tokens
			if not None and ignores maximum size.
		"""
		if wordlist is None:
			# work with the supplied tokens. extract frequencies.

			# gets frequency count
			c = self._get_frequency_count(tokens)

			if minimum_occurrences is None:
				minimum_occurrences = 1

			words = [key for key, number in c.most_common()
					 if number >= minimum_occurrences and key]

			if size is not None and size < len(words):
				words = words[:size]
		else:
			# using ordered dict as an ordered set
			# (we need to keep the order and eliminate duplicates)
			words = [word.lower().translate(WordDictionary.number_transformation) for word in wordlist]
			values = [None] * len(words)
			words = OD(zip(words, values)).keys()
			#for c in words:
			#	print c.encode('utf-8')

		# verifies the maximum size
		if size is None:
			size = len(words)

		# set all words in the dictionary
		for word, num in zip(words, range(size)):
			self[word] = num

		# if the given words include one of the the rare or padding symbols, don't replace it
		# 사전 생성 시에 가장 뒤에 padding, rare를 생성
		special_symbols = [WordDictionary.rare.lower(),
						   WordDictionary.padding_left.lower(),
						   WordDictionary.padding_right.lower()]

		for symbol in special_symbols:
			if symbol not in words:
				self[symbol] = size
				size += 1

		self.check()

	@classmethod
	def init_from_wordlist(cls, wordlist):
		"""
		Initializes the WordDictionary instance with a list of words, independently from their
		frequencies. Every word in the list gets an entry.
		"""
		return cls(None, wordlist=wordlist)

	@classmethod
	def init_empty(cls):
		"""
		Initializes an empty Word Dictionary.
		"""
		return cls([[]])

	def save(self, filename):
		"""
		Saves the word dictionary to the given file as a list of word types.

		Special words (paddings and rare) are also included.
		"""
		sorted_words = sorted(self, key=self.get)
		text = '\n'.join(sorted_words)
		with open(filename, 'w') as f:
			f.write(text)

	@classmethod
	def load(cls, filename):
		"""
		Loads a WordDictionary object from a vocabulary file.
		"""
		words = []
		with open(filename, 'r') as f:
			for word in f:
				word = word.strip()
				if word:
					words.append(word)

		return cls.init_from_wordlist(words)

	def _get_frequency_count(self, token_list):
		"""
		Returns a token counter for tokens in token_list.

		:param token_list: Either a list of tokens (as strings) or a list
			of lists of tokens.
		"""
		if type(token_list[0]) == list:
			c = Counter(t.lower().translate(WordDictionary.number_transformation)
						for sent in token_list for t in sent)
		else:
			c = Counter(t.lower().translate(WordDictionary.number_transformation)
						for t in token_list)
		return c


	def update_tokens(self, tokens, size=None, minimum_occurrences=1, freqs=None):
		"""
		Updates the dictionary, adding more types until size is reached.

		:param freqs: a dictionary providing a token count.
		"""
		if freqs is None:
			freqs = self._get_frequency_count(tokens)

		if size is None or size == 0:
			# size None or 0 means no size limit
			size = len(freqs)

		if self.num_tokens >= size:
			return
		else:
			size_diff = size - self.num_tokens

		# a new version of freqs with only tokens not present in the dictionary
		# and above minimum frequency
		candidate_tokens = dict((token, freqs[token])
								for token in freqs
								if token not in self and freqs[token] >= minimum_occurrences)

		# order the types from the most frequent to the least
		new_tokens = sorted(candidate_tokens, key=lambda x: candidate_tokens[x], reverse=True)

		next_value = len(self)
		for token in new_tokens:
			self[token] = next_value
			next_value += 1
			size_diff -= 1
			if size_diff == 0:
				break

		self.check()

	def __contains__(self, key):
		"""
		Overrides the "in" operator. Case insensitive.
		"""
		transformed = key.lower().translate(WordDictionary.number_transformation)
		return super(WordDictionary, self).__contains__(transformed)

	def __setitem__(self, key, value):
		"""
		Overrides the [] write operator. It converts every key to lower case
		before assignment.
		"""
		transformed = key.lower().translate(WordDictionary.number_transformation)
		super(WordDictionary, self).__setitem__(transformed, value)

	def __getitem__(self, key):
		"""
		Overrides the [] read operator.

		Three differences from the original:
		1) when given a word without an entry, it returns the value for the *UNKNOWN* key.
		2) all entries are converted to lower case before verification.
		3) digits are mapped to 9
		"""
		# faster than regexp
		transformed = key.lower().translate(WordDictionary.number_transformation)
		return super(WordDictionary, self).get(transformed, self.index_rare)

	def get(self, key):
		"""
		Overrides the dictionary get method, so when given a word without an entry, it returns
		the value for the *rare* key. Note that it is not possible to supply a default value as
		in the dict class.
		"""
		# faster than regexp
		#print (key, type(key))
		transformed = key.lower().translate(WordDictionary.number_transformation)
		#print key, transformed
		return super(WordDictionary, self).get(transformed, self.index_rare)

	def check(self):
		"""
		Checks the internal structure of the dictionary and makes necessary adjustments,
		such as updating num_tokens.
		"""
		# since WordDictionary overrides __get__, we use the super call
		# (the WordDictionary __get__ fails when self.index_rare is not set)
		key = WordDictionary.rare.lower()
		self.index_rare = super(WordDictionary, self).get(key)

		self.index_padding_left = self[WordDictionary.padding_left]
		self.index_padding_right = self[WordDictionary.padding_right]
		self.num_tokens = len(self)

	def get_words(self, indices):
		"""
		Returns the words represented by a sequence of indices.
		"""
		words = [w for w in self if self[w] in indices]
		return words

	def get_indices(self, words):
		"""
		Returns the indices corresponding to a sequence of tokens.
		"""
		indices = [self[w] for w in words]
		return indices



# -*- coding: utf-8 -*-

"""
Class for dealing with POS data.
"""

from ..reader import TaggerReader

class ConllNER(object):
	"""
	Dummy class for storing column positions in a conll file.
	"""
	id = 0
	word = 1
	pos = 2
	ner = 3
	link = 4
	SEP = '\t'

class NERReader(TaggerReader):
	"""
	This class reads data from a POS corpus and turns it into a format
	readable by the neural network for the POS tagging task.
	"""
	
	def __init__(self, md=None, filename=None, load_dictionaries=True):
		"""
		Constructor
		"""
		self.rare_tag = None
		self.sentences = []
		if filename is not None:
			try:
				self._read_plain(filename)
			except:
				self._read_conll(filename)
		
		super(NERReader, self).__init__(md, load_dictionaries=load_dictionaries)
		
	@property
	def task(self):
		"""
		Abstract Base Class (ABC) attribute.
		"""
		return 'ner'
	
	def _read_plain(self, filename):
		"""
		Read data from a "plain" file, with one sentence per line, each token
		as token_tag.
		"""
		self.sentences = []
		with open(filename, 'rt') as f:
			for line in f:
				#line = unicode(line, 'utf-8')
				items = line.strip().split()
				sentence = []
				for item in items:
					token, tag = item.rsplit('_', 1)
					sentence.append((token, tag))
					
				self.sentences.append(sentence)
	
	def _read_conll(self, filename):
		"""
		Read data from a CoNLL formatted file. It expects at least 4 columns:
		id, surface word, lemma (ignored, may be anything) 
		and the POS tag.
		"""
		self.sentences = []
		sentence = []
		with open(filename, 'rt') as f:
			for line in f:
				line = line.strip()
				if line == '':
					if len(sentence) > 0:
						self.sentences.append(sentence)
					sentence = [] # 문장 끝나고 빈 라인이 여러개 있는 것에 대비 
					continue
				
				fields = line.split(ConllNER.SEP)
				word = fields[ConllNER.word]
				pos = fields[ConllNER.pos]
				ner = fields[ConllNER.ner]
				link = fields[ConllNER.link]
				sentence.append((word, ner))
		
		if len(sentence) > 0:
			self.sentences.append(sentence)

# backwards compatibility
MacMorphoReader = NERReader



from nltk.translate.bleu_score import *
from nltkor.tokenize import Ko_tokenize
import numpy as np
import torch
import time
import math

def bleu_tensor(reference,candidate,n=0, smoothing_function=None):
    if n: weights = tuple(1 if i == n-1 else 0 for i in range(4))
    else: weights = (0.25, 0.25, 0.25, 0.25)

 
  
    reference=reference.unsqueeze(1)
    reference=reference.numpy()
    candidate=candidate.numpy()
    return torch.tensor(corpus_bleu(reference,candidate,weights,smoothing_function=smoothing_function))






# Natural Language Toolkit for Korean: NLTKor's very own tokenizer.
#
# Copyright (C) 2001-2020 NLTKor Project
# Author:
# URL: <http://>
# For license information, see LICENSE.TXT


import re

#for Korean

class Enum(object):
    def __init__(self, names):
        for value, name in enumerate(names.split()): setattr(self, name, value)

class Ko_tokenize():

    def word(target, encoding='utf8'):
        """ Word Tokenizer

        단어 단위로 Tokenizing 한다.

        인자값 목록 (모든 변수가 반드시 필요):

        target : Tokenizing 하고자 하는 대상 문자열

        결과값 : 토크나이징 결과를 list 자료형으로 넘김

        """
        isHangulSyllables = lambda x: unicodedata.name(x).find("HANGUL SYLLABLE") == 0
        isHanjaSyllables = lambda x: unicodedata.name(x).find("CJK") == 0
        isNumber = lambda x: unicodedata.name(x).find("FULLWIDTH DIGIT") == 0 or unicodedata.name(x).find("DIGIT") == 0
        isAlphabet = lambda x: unicodedata.name(x).find("FULLWIDTH LATIN") == 0 or unicodedata.name(x).find("LATIN") == 0
        isAlphabet_Connection = lambda x: x in (".", "-", "_", "|")
        isNumber_Connection = lambda x: x in (".", ",")
        isPunctuation = lambda x: unicodedata.category(x)[0] == "P"
        isSymbol = lambda x: unicodedata.category(x)[0] == "S"
        getCategory = lambda x: unicodedata.category(x)

        TYPE = Enum("UNKNOWN SYMBOL NUMBER PUNCTUATION ALPHABET HANJA HANGUL")

        buf = str()
        type_prev = 0
        type_cur = 0

        if type(target) == str:
            target = target

        for i in range(len(target)):
            ch = target[i]
            ca = str()
            try:
                if isHangulSyllables(ch): type_cur = TYPE.HANGUL
                elif isHanjaSyllables(ch): type_cur = TYPE.HANJA
                elif isNumber(ch): type_cur = TYPE.NUMBER
                elif isAlphabet(ch): type_cur = TYPE.ALPHABET
                elif isAlphabet_Connection(ch) and type_prev == TYPE.ALPHABET:
                    if i+1 < len(target) and not isAlphabet(target[i+1]): type_cur = TYPE.SYMBOL
                    else: type_cur = TYPE.ALPHABET
                elif isNumber_Connection(ch) and type_prev == TYPE.NUMBER:
                    if i+1 < len(target) and not isNumber(target[i+1]): type_cur = TYPE.SYMBOL
                    elif i+1 == len(target): type_cur = TYPE.SYMBOL
                    else: type_cur = TYPE.NUMBER
                elif isPunctuation(ch): type_cur = TYPE.PUNCTUATION
                elif isSymbol(ch): type_cur = TYPE.SYMBOL
                else: type_cur = TYPE.UNKNOWN
                ca = getCategory(ch)
            except:
                type_cur = TYPE.UNKNOWN
            if type_cur == TYPE.PUNCTUATION:
                if ca in ("Ps", "Pe"): buf += " "
                elif i >= 0 and i < len(target) and target[i-1] != target[i]: buf += " "
            elif type_cur != type_prev: buf += " "
            buf += ch
            type_prev = type_cur
        return buf.split()



    def syllable(text,blank=False):
        """
        음절 토크나이저

        음절단위로 tokenizing

        박찬양
        """
        emjeol_list = list()
        for emjeol in text:

          if blank and (emjeol not in ['\n']):
           emjeol_list.append(emjeol)

          elif emjeol not in [' ', '\n']:
           emjeol_list.append(emjeol)

        return emjeol_list


    def sentence(text):
        """
        문장 토크나이저

        문장단위로 tokenizing
        """
        txt=text.replace("\n"," ")
        p=re.compile(r'(?<!\w\.\w.)(?<=\.|\?|\!)\s').split(txt)
        result=[]
        for tmp in p:
            if (tmp == ' ' or tmp== ''):
                continue
            else:result.append(tmp.strip(" "))

        return result



# Natural Language Toolkit: Metrics
#
# Copyright (C) 2001-2020 NLTK Project
# Author: Steven Bird <stevenbird1@gmail.com>
#         Edward Loper <edloper@gmail.com>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
#

"""
NLTKor Metrics

Classes and methods for scoring processing modules.
"""
"""
from nltk.metrics.scores import (
    accuracy,
    precision,
    recall,
    f_measure,
    log_likelihood,
    approxrand,
)
from nltk.metrics.confusionmatrix import ConfusionMatrix
from nltk.metrics.distance import (
    edit_distance,
    edit_distance_align,
    binary_distance,
    jaccard_distance,
    masi_distance,
    interval_distance,
    custom_distance,
    presence,
    fractional_presence,
)
from nltk.metrics.paice import Paice
from nltk.metrics.segmentation import windowdiff, ghd, pk
from nltk.metrics.agreement import AnnotationTask
from nltk.metrics.association import (
    NgramAssocMeasures,
    BigramAssocMeasures,
    TrigramAssocMeasures,
    QuadgramAssocMeasures,
    ContingencyMeasures,
)
from nltk.metrics.spearman import (
    spearman_correlation,
    ranks_from_sequence,
    ranks_from_scores,
)
from nltk.metrics.aline import align
from nltkor.metrics.eval import StringMetric
"""
from nltkor.metrics.classical import DefaultMetric
from nltkor.metrics.entment import EntMent
from nltkor.metrics.bleu_tensor import *
#DefaultMetric = lazy_import.lazy_callable("nltkor.metrics.classical.DefaultMetric")
#Mauve = lazy_import.lazy_callable("nltkor.metrics.mauve.Mauve")
from nltkor.metrics.mauve import Mauve
#BERTScore = lazy_import.lazy_callable("nltkor.metrics.bertscore.BERTScore")
from .bertscore import BERTScore
#BARTScore = lazy_import.lazy_callable("nltkor.metrics.bartscore.BARTScore")
from .bartscore import BARTScore



# -*- coding: utf-8 -*-

"""
Taggers wrapping the neural networks.
"""

import logging
#from os import major
import numpy as np
import re
from nltkor import etc

from . import utils
from . import config
from . import attributes
from .metadata import Metadata
from .pos import POSReader
from .ner import NERReader
from .wsd import WSDReader
from .srl import SRLReader
from .parse import DependencyReader
import sys
sys.path.append("libs/")
from .network import Network, ConvolutionalNetwork, ConvolutionalDependencyNetwork


def load_network(md):
		"""
		Loads the network from the default file and returns it.
		"""
		logger = logging.getLogger("Logger")
		is_srl = md.task == 'srl'

		logger.info('Loading network')
		if is_srl :
				net_class = ConvolutionalNetwork
		elif md.task.endswith('dependency'):
				net_class = ConvolutionalDependencyNetwork
		else:
				net_class = Network

		nn = net_class.load_from_file(md.paths[md.network])

		logger.info('Done')
		return nn


def create_reader(md, gold_file=None):
		"""
		Creates a TextReader object for the given task and loads its dictionary.
		:param md: a metadata object describing the task
		:param gold_file: path to a file with gold standard data, if
				the reader will be used for testing.
		"""
		logger = logging.getLogger('Logger')
		logger.info('Loading text reader...')

		if md.task == 'pos':
				tr = POSReader(md, filename=gold_file)

		elif md.task == 'ner':
				tr = NERReader(md, filename=gold_file)

		elif md.task == 'wsd':
				tr = WSDReader(md, filename=gold_file)

		elif 'dependency' in md.task:
				labeled = md.task.startswith('labeled')
				tr = DependencyReader(md, filename=gold_file, labeled=labeled)

		elif md.task.startswith('srl'):
				tr = SRLReader(md, filename=gold_file)

		else:
				raise ValueError("Unknown task: %s" % md.task)

		logger.info('Done')
		return tr

def _group_arguments(tokens, predicate_positions, arg_tokens, labels):
		"""
		Groups words pertaining to each argument and returns a dictionary for each predicate.
		"""
		print(tokens, predicate_positions, arg_tokens, labels)
		arg_structs = []

		for predicate_position, pred_arg_tokens, pred_labels in zip(predicate_positions,
																																 arg_tokens,
																																 labels):
				structure = {}

				for tag, arg_token in zip(pred_labels, pred_arg_tokens):
					#argument_tokens = [token]
					#tag = pred_labels.pop(0)
					structure[tag] = [arg_token]

				predicate = tokens[predicate_position-1]
				arg_structs.append((predicate, structure))

		return arg_structs


class SRLAnnotatedSentence(object):
		"""
		Class storing a sentence with annotated semantic roles.

		It stores a list with the sentence tokens, called `tokens`, and a list of tuples
		in the format `(predicate, arg_strucutres)`. Each `arg_structure` is a dict mapping
		semantic roles to the words that constitute it. This is used instead of a two-level
		dictionary because one sentence may have more than one occurrence of the same
		predicate.

		This class is used only for storing data.
		"""

		def __init__(self, tokens, arg_structures):
				"""
				Creates an instance of a sentence with SRL data.

				:param tokens: a list of strings
				:param arg_structures: a list of tuples in the format (predicate, mapping).
						Each predicate is a string and each mapping is a dictionary mapping role labels
						to the words that constitute it.
				"""
				self.tokens = tokens
				self.arg_structures = arg_structures

class ParsedSentence(object):
		"""
		Class for storing a sentence with dependency parsing annotation.

		It stores a list of tokens, the dependency heads, dependency labels and POS tags
		if the parser used them. Dependency heads are the index of the head of each
		token, and -1 means a dependency to the root.
		"""
		def __init__(self, tokens, heads, labels, pos=None):
				"""
				Constructor.

				:param tokens: list of strings
				:param heads: list of integers (-1 means dependency to root, others are token indices)
				:param labels: list of strings
				:param pos: None or list of strings
				"""
				self.tokens = tokens
				self.heads = heads
				self.labels = labels
				self.pos = pos

		def __len__(self):
				return len(self.tokens)

		def to_conll_list(self):
				"""
				Return a list representation of the sentence in CoNLL X format.

				Each line has:
				[number starting from 1] token _ POS POS _ head label

				Token numbers start from 1, root is referred as 0.
				POS is only available if the original parser used it.
				"""
				tokenL = []
				headL = []
				labelL = []
				posL = []
				for i in range(len(self.tokens)):
						tokenL.append(self.tokens[i])
						headL.append(self.heads[i] + 1)
						labelL.append(self.labels[i])
						posL.append(self.pos[i])

				return tokenL, posL, labelL, headL

		def to_conll(self):
				"""
				Return a string representation of the sentence in CoNLL X format.

				Each line has:
				[number starting from 1] token _ POS POS _ head label

				Token numbers start from 1, root is referred as 0.
				POS is only available if the original parser used it.
				"""
				result = []
				for i in range(len(self.tokens)):
						token = self.tokens[i]
						head = self.heads[i] + 1
						label = self.labels[i]
						pos = self.pos[i] if self.pos else '_'

						#line = u'{id}\t{token}\t_\t{pos}\t{pos}\t_\t{head}\t{label}'
						#result.append(line.format(id=i+1, pos=pos, head=head, label=label, token=token))
						line = u'{id}\t{token}\t{head}\t{label}'
						result.append(line.format(id=i+1, head=head, label=label, token=token))

				return '\n'.join(result)


class Tagger(object):
		"""
		Base class for taggers. It should not be instantiated.
		"""
		def __init__(self, data_dir=None):
				"""Creates a tagger and loads data preemptively"""
				asrt_msg = "espresso data directory is not set. \
If you don't have the trained models, download them from http://air.cwnu.ac.kr/espresso/models.html"
				if data_dir is None:
						assert config.data_dir is not None, asrt_msg
						self.paths = config.FILES
				else:
						self.paths = config.get_config_paths(data_dir)

				self.data_dir = data_dir
				self._load_data()

		def _load_data(self):
				"""Implemented by subclasses"""
				pass


class SRLTagger(Tagger):
		"""
		An SRLTagger loads the models and performs SRL on text.

		It works on three stages: predicate identification, argument detection and
		argument classification.
		"""

		def _load_data(self):
				"""Loads data for SRL"""
				md_srl = Metadata.load_from_file('srl', self.paths)
				self.nn = load_network(md_srl)
				self.reader = create_reader(md_srl)
				self.reader.create_converter()
				self.itd = self.reader.get_inverse_tag_dictionary()

				self.parser = DependencyParser(self.data_dir)


		def find_predicates(self, tokens):
				"""
				Finds out which tokens are predicates.

				:param tokens: a list of attribute.Token elements
				:returns: the indices of predicate tokens
				"""
				answer = []
				for i, token in enumerate(tokens):
					if token[0] == 'V' and tokens[i-1][0] != 'V': answer.append(i+1)
				return np.array(answer)

		def find_arguments(self, token_obj, predL, headL, relL):
				"""
				Finds out which tokens are predicates.

				:param tokens: a list of attribute.Token elements
				:returns: the indices of predicate tokens
				"""
				answer_token = []; answer = []
				for p in predL:
					pred_arg_token = []; pred_arg = []
					for j, h in enumerate(headL):
						if p == h and relL[j][0] == 'N':
							pred_arg_token.append(token_obj[j])
							pred_arg.append(np.array([j, j]))

					#TODO
					# predicate의 header
					#if headL[p-1] != 0: # 마지막 제외
					#	pred_arg_token.append(token_obj[headL[p-1]-1])
					#	pred_arg.append(np.array([headL[p-1]-1, headL[p-1]]))

					answer_token.append(pred_arg_token)
					answer.append(pred_arg)
				#print(answer_token)
				#print(answer)
				return answer_token, answer

		def tag(self, text, use_sent_tokenizer=True, mode='standard'):
				"""
				Runs the SRL process on the given text.

				:param text: unicode or str encoded in utf-8.
				:param no_repeats: whether to prevent repeated argument labels
				:returns: a list of SRLAnnotatedSentence objects
				"""
				tokens = utils.tokenize(text)
				result = []
				for sent in tokens:
						tagged = self.tag_sentence(sent)
						result.append(tagged)

				return result

		def tag_sentence(self, tokens, no_repeats=False):
				"""
				Runs the SRL process on the given tokens.

				:param tokens: a list of tokens (as strings)
				:param no_repeats: whether to prevent repeated argument labels
				:returns: a list of lists (one list for each sentence). Sentences have tuples
						(all_tokens, predicate, arg_structure), where arg_structure is a dictionary
						mapping argument labels to the words it includes.
				"""
				# 구문분석
				parsed = self.parser.parse_sentence(tokens)
				wordL, posL, relL, headL =  parsed.to_conll_list()
				tokens_obj = []
				for w, p, r in zip(wordL, posL, relL):
					hm, hp, tm, tp = p
					token = attributes.Token(w, hm, hp, tm, tp, r)
					tokens_obj.append(token)

				converted_class = np.array([self.reader.converter.convert(t)
																		for t in tokens_obj])
				pred_positions = self.find_predicates(relL)

				arg_tokens, arg_limits = self.find_arguments(wordL, pred_positions, headL, relL)
				print(arg_tokens)
				print(pred_positions)
				print(arg_limits)

				# now, argument classification
				answers = self.nn.tag_sentence(converted_class,
																								pred_positions, arg_limits,
																								allow_repeats=not no_repeats)
				labels = [[self.itd[x] for x in pred_answer]
										 for pred_answer in answers]

				structures = _group_arguments(wordL, pred_positions, arg_tokens, labels)
				return SRLAnnotatedSentence(wordL, structures)

class DependencyParser(Tagger):
		"""A Dependency Parser based on a neural network tagger."""

		def __init__(self, *args, **kwargs):
				"""
				Set the data directory for the POS tagger, if one is used,
				and call the parent constructor.
				"""
				super(DependencyParser, self).__init__(*args, **kwargs)

		def _load_data(self):
				"""Loads data for Dependency Parsing"""
				md_udep = Metadata.load_from_file('unlabeled_dependency', paths=self.paths)
				self.unlabeled_nn = load_network(md_udep)
				self.unlabeled_reader = create_reader(md_udep)

				md_ldep = Metadata.load_from_file('labeled_dependency', paths=self.paths)
				self.labeled_nn = load_network(md_ldep)
				self.labeled_reader = create_reader(md_ldep)
				self.itd = self.labeled_reader.get_inverse_tag_dictionary()

				self.use_pos = md_udep.use_pos or md_ldep.use_pos
				if self.use_pos:
						self.pos_tagger = POSTagger(self.data_dir)

		def parse(self, text):
				"""
				Split the given text into sentences and determines their
				dependency trees. If you want to provide your own tokenized
				text, use `parse_sentence` instead.

				:param text: a string
				:returns: a list of ParsedSentence's
				"""
				sentences = utils.tokenize(text)
				result = []
				for sent in sentences:
						parsed = self.parse_sentence(sent)
						result.append(parsed)

				return result

		def tag_tokens(self, tokens):
			"""
			Parse the given sentence. This function is just an alias for
			`parse_sentence`.
			"""
			return self.parse_sentence(tokens)

		def parse_sentence(self, tokens):
			"""
			Parse the given sentence. It must be already tokenized; if you
			want nlpnet to tokenize the text, use the method `parse` instead.

			:param tokens: a list of strings (sentences)
			:return: a ParsedSentence instance
			"""
			original_tokens = tokens
			udep_tokens_obj = []
			ldep_tokens_obj = []

			# if the parser uses POS a feature, have a tagger tag it first
			if self.use_pos:
				eojeols, eojeol_features = self.pos_tagger.tag_tokens(tokens, mode='eojeol')
				#print("**", eojeols)
				#print(eojeol_features)
			#print(tokens, eojeols)

			for word, feature in zip(eojeols, eojeol_features):
				m_h, t_h, m_t, t_t = feature
				#udep_tokens_obj.append(attributes.Token(word, morph_h=m_h, pos_h=t_h, morph_t=m_t, pos_t=t_t))
				udep_tokens_obj.append(attributes.Token(word, pos_h=t_h, morph_t=m_t, pos_t=t_t))
				ldep_tokens_obj.append(attributes.Token(word, pos_h=t_h, morph_t=m_t, pos_t=t_t))

			converted_tokens = self.unlabeled_reader.codify_sentence(udep_tokens_obj)
			#print(converted_tokens)
			heads = self.unlabeled_nn.tag_sentence(converted_tokens)
			#print(heads)

			# the root is returned having a value == len(sentence)
			root = heads.argmax()
			heads[root] = root

			converted_tokens = self.labeled_reader.codify_sentence(ldep_tokens_obj)
			label_codes = self.labeled_nn.tag_sentence(converted_tokens, heads)
			labels = [self.itd[code] for code in label_codes]
			#print(label_codes)
			#print(labels)

			# to the final answer, signal the root with -1
			heads[root] = -1
			pos_tags = eojeol_features if self.use_pos else None
			#pos_tags = zip(*tokens)[1] if self.use_pos else None

			parsed = ParsedSentence(eojeols, heads, labels, pos_tags)
			#parsed = ParsedSentence(original_tokens, heads, labels, pos_tags)
			return parsed

		def tag(self, text, use_sent_tokenizer=True, mode='eojeol'):
			"""
			Parse the given text. This is just an alias for the
			`parse` method.
			"""
			return self.parse(text)


class WSDTagger(Tagger):
		"""A WSDTagger loads the models and performs WSD tagging on text."""

		def _load_data(self):
				"""Loads data for WSD"""
				md_wsd = Metadata.load_from_file('wsd', self.paths)
				self.nn = load_network(md_wsd)
				self.reader = create_reader(md_wsd)
				self.reader.create_converter()
				self.itd = self.reader.get_inverse_tag_dictionary()
				#self.morph_lexicon = self.reader.morph_lexicon					# user lexicon
				#self.co_lexicon = self.reader.co_lexicon
				#self.prob_dict = self.reader.prob_dict
				self.pos_tagger = POSTagger(self.data_dir)

		def tag(self, text, use_sent_tokenizer=True, mode='standard'):
				"""
				Tags the given text.

				:param text: a string or unicode object. Strings assumed to be utf-8
				:returns: a list of lists (sentences with tokens).
						Each sentence has (token, tag) tuples.
				"""
				tokens = utils.tokenize(text)
				result = []
				for sent in tokens:
					tagged = self.tag_sentence(sent)
					result.append(tagged)

				return result

		def tag_sentence(self, tokens):
			"""
			Tags a given list of tokens.

			Tokens should be produced with the espresso tokenizer in order to
			match the entries in the vocabulary. If you have non-tokenized text,
			use NERTagger.tag(text).

			:param tokens: a list of strings
			:returns: a list of strings (morphs, tags)
			"""
			pos_tagged = self.pos_tagger.tag_tokens(tokens)

			pos_tagged = filter(lambda x : x != (' ', 'SP'), pos_tagged)
			unzipped_pos_tagged = zip(*pos_tagged)
			morphs, morph_pos_tags = list(unzipped_pos_tagged)
			#print(morphs, morph_pos_tags)

			converter = self.reader.converter
			converted_tokens = np.array([converter.convert(token) for token in morphs])
			#print("0", converted_tokens)

			answer = self.nn.tag_sentence(converted_tokens)
			tags = [self.itd[tag] for tag in answer]				# 번호를 수로 표현

			#print("1", morphs, tags)

			return zip(morphs, tags)


class NERTagger(Tagger):
		"""A NERTagger loads the models and performs NER tagging on text."""

		def _load_data(self):
				"""Loads data for NER"""
				md_ner = Metadata.load_from_file('ner', self.paths)
				self.nn = load_network(md_ner)
				self.reader = create_reader(md_ner)
				self.reader.create_converter()
				self.itd = self.reader.get_inverse_tag_dictionary()
				#self.morph_lexicon = self.reader.morph_lexicon					# user lexicon
				#self.co_lexicon = self.reader.co_lexicon
				#self.prob_dict = self.reader.prob_dict
				self.pos_tagger = POSTagger(self.data_dir)

		def tag(self, text, use_sent_tokenizer=True, mode='standard'):
				"""
				Tags the given text.

				:param text: a string or unicode object. Strings assumed to be utf-8
				:returns: a list of lists (sentences with tokens).
						Each sentence has (token, tag) tuples.
				"""
				tokens = utils.tokenize(text)
				result = []
				for sent in tokens:
					tagged = self.tag_sentence(sent)
					result.append(tagged)

				return result

		def tag_sentence(self, tokens):
			"""
			Tags a given list of tokens.

			Tokens should be produced with the espresso tokenizer in order to
			match the entries in the vocabulary. If you have non-tokenized text,
			use NERTagger.tag(text).

			:param tokens: a list of strings
			:returns: a list of strings (morphs, tags)
			"""
			pos_tagged = self.pos_tagger.tag_tokens(tokens)

			pos_tagged = filter(lambda x : x != (' ', 'SP'), pos_tagged) # 공백 제거
			unzipped_pos_tagged = zip(*pos_tagged)
			morphs, morph_pos_tags = list(unzipped_pos_tagged)
			#print(morphs, morph_pos_tags)

			converter = self.reader.converter
			converted_tokens = np.array([converter.convert(token) for token in morphs])
			#print("0", converted_tokens)

			answer = self.nn.tag_sentence(converted_tokens)
			tags = [self.itd[tag] for tag in answer]				# 번호를 수로 표현

			#print("1", morphs, tags)

			return zip(morphs, tags)


class POSTagger(Tagger):
		"""A POSTagger loads the models and performs POS tagging on text."""

		def _load_data(self):
				"""Loads data for POS"""
				md = Metadata.load_from_file('pos', self.paths)
				self.nn = load_network(md)
				self.reader = create_reader(md)
				self.reader.create_converter()
				self.itd = self.reader.get_inverse_tag_dictionary()
				self.morph_lexicon = self.reader.morph_lexicon					# user lexicon
				self.co_lexicon = self.reader.co_lexicon
				self.prob_dict = self.reader.prob_dict

		def tag(self, text, use_sent_tokenizer=True, mode="standard"):
				"""
				Tags the given text.

				:param text: a string or unicode object. Strings assumed to be utf-8
				:param mode:  [standard, eumjeol, eojeol]. "eumjeol" does not lemmatize,
						"eojeol" includes NN+XV
				:returns: a list of lists (sentences with tokens).
						Each sentence has (token, tag) tuples.
				"""
				tokens = utils.tokenize(text, use_sent_tokenizer) # 문장 단위 + 음절
				result = []
				for sent in tokens:
						tagged = self.tag_tokens(sent, mode)
						result.append(tagged)

				return result

		def tag_tokens(self, tokens, mode="standard"):
				"""
				Tags a given list of tokens.

				Tokens should be produced with the espresso tokenizer in order to
				match the entries in the vocabulary. If you have non-tokenized text,
				use POSTagger.tag(text).

				:param tokens: a list of strings
				:param mode: [standard, eumjeol, verb]. "eumjeol" does not lemmatize,
						"verb" includes NN+XV
				:returns: a list of strings (the tags)
				"""
				converter = self.reader.converter			# 클래스 지정
				converted_tokens = np.array([converter.convert('*space*') if token==' ' else converter.convert(token)
																		 for token in tokens])
				#print("0", converted_tokens)

				answer = self.nn.tag_sentence(converted_tokens)
				tags = [self.itd[tag] for tag in answer]				# 번호를 문자로 변환

				if mode == 'eojeol':
					eojeols, eojeol_features = self.get_eojeol_tokens(tokens, tags, mode)
					return eojeols, eojeol_features
				else:
					morphs, morph_tags = self.get_morph_tokens(tokens, tags, mode)
					return zip(morphs, morph_tags)


		def _get_morph_tokens(self, tokens, tags):
			"""
			공백으로 형태소 분리.

			:param tokens: a list of strings
			:param tags: a list of tags of each string
			:return: a list of (morph, tag)
			"""
			#print(utils.get_word(self.morph_lexicon, tokens, tags, True))
			# 기분석 사전 검색
			tokens, tags = utils.get_word(self.morph_lexicon, tokens, tags, True)
			#print(tokens)
			#print(tags)
			morphs = [''.join(tokens[0]) if isinstance(tokens[0], list) else tokens[0]]
			morph_tags = [(lambda x: 'MA' if x == 'MS' else x)\
						((lambda x: 'NN' if x == 'NS' else x)(tags[0]))]
			for idx in range(1,len(tokens)):
				if (tags[idx-1]=='NS' and tags[idx]=='NN') \
						or (tags[idx-1]=='MS' and tags[idx]=='MA'):
					morphs.append(morphs.pop()+(''.join(tokens[idx]) if isinstance(tokens[idx], list) else tokens[idx]))
				elif tags[idx-1] != tags[idx] or tags[idx] == 'SY':
					morphs.append(''.join(tokens[idx]) if isinstance(tokens[idx], list) else tokens[idx])
					morph_tags.append((lambda x: 'MA' if x == 'MS' else x)\
						((lambda x: 'NN' if x == 'NS' else x)(tags[idx])))
				else:
					morphs.append(morphs.pop()+(''.join(tokens[idx]) if isinstance(tokens[idx], list) else tokens[idx]))

			return morphs, morph_tags


		def get_eumjeol_tokens(self, tokens, tags):
			"""
			음절 토큰으로 처리.
			'CO'를 앞 형태소에 붙이고 품사는 앞의 것을 따름
			새로운 -> 새/VB+로운/CO -> 새로운/VB

			:param tokens: a list of strings
			:param tags: a list of tags of each string
			:return: a list of (eumjeol, tag)
			"""
			eumjeol = []
			eumjeol_tags = []
			#print(tokens)
			#print(tags)
			for idx in range(0, len(tokens)):
				if idx>0 and (tags[idx]=='CO' and \
						tags[idx-1]!='SP' and tags[idx-1][1]!='N'):
					eumjeol.append(eumjeol.pop()+(''.join(tokens[idx]) if isinstance(tokens[idx], list) else tokens[idx]))
				elif idx>0 and (tags[idx]=='CO' and \
						tags[idx-1]!='SP' and tags[idx-1][1]=='N'):
					eumjeol.append(tokens[idx])
					eumjeol_tags.append('XV')
				elif tags[idx] =='CO':
					eumjeol.append(tokens[idx])
					eumjeol_tags.append('VB')
				else:
					eumjeol.append(tokens[idx])
					eumjeol_tags.append(tags[idx])
			#print(eumjeol)
			#print(eumjeol_tags)

			return eumjeol, eumjeol_tags

		def get_eojeol(self, tokens, tags):
			"""
			음절 토큰으로 처리.
			'CO'를 앞 형태소에 붙이고 품사는 앞의 것을 따름
			새로운 -> 새/VB+로운/CO -> 새로운/VB

			:param tokens: a list of strings
			:param tags: a list of tags of each string
			:return: a list of (eumjeol, tag)
			"""
			eojeols = []
			eumjeol = []
			#print(tokens)
			for t in tokens:
				if t == ' ':
					eojeols.append(''.join(eumjeol))
					eumjeol = []
				else:
					eumjeol.append(t)
			eojeols.append(''.join(eumjeol))
			#print(eojeols)

			return eojeols



		def get_morph_tokens(self, tokens, tags, mode="standard"):
			"""
			combine eumjeol to morph

			param tokens: eumjeol token list
			param tags: pos tag list of each token
			"""
			_morphs, _morph_tags = self._get_morph_tokens(tokens, tags)
			#print('2---', morphs, morph_tags) # 원형 복원 전

			if mode=='eumjeol':
				eumjeols, eumjeol_tags = self.get_eumjeol_tokens(_morphs, _morph_tags)
				return eumjeols, eumjeol_tags

			# 'CO', 축약어 처리
			morphs, morph_tags = self.handling_abbrs(_morphs, _morph_tags)
			#print("3", morphs, morph_tags)  # 원형복원

			if mode=='eojeol':
				eojeols = self.get_eojeol(_morphs, _morph_tags)
				return eojeols, morphs, morph_tags
			return morphs, morph_tags

		def handling_abbrs(self, _morphs, _tags):
			'''
			CO tag와 축약을 처리한다.
			'''
			morphs = []
			morph_tags = []
			#print(_morphs, _tags, flush=True)

			for i, t in enumerate(_tags):
				if t == 'CO':
					prev_morph = _morphs[i-1] if i > 0 else 'BOS'
					prev_tag = _tags[i-1] if i > 0 else 'BOS'
					next_morph = _morphs[i+1] if i < len(_tags)-1 else 'EOS'
					next_tag   = _tags[i+1] if i < len(_tags)-1 else 'EOS'
					if _tags[i-1] in ['MM']:
						morph_tags[-1] = 'NN'; _tags[i-1] = 'NN'; prev_tag='NN'  # 나중 제거할 것
					morphs, morph_tags =  self.handling_co_tags(morphs, morph_tags, \
						prev_morph, prev_tag, _morphs[i], _tags[i], next_morph, next_tag)
				elif i > 0:
					morphs, morph_tags = self.handling_others(morphs, morph_tags, _morphs[i], _tags[i])
				else:
					if _morphs[i] == ' ': t = 'SP'
					morphs.append(_morphs[i])
					morph_tags.append(t)
			return morphs, morph_tags


		def handling_others(self, morphs, morph_tags, morph, tag):
			'''
			다른 형태소들의 처리를 담당한다. 여기에는
			1. 'ㄴ,ㄹ, ㅁ, ㅂ, ㅆ' 의 축약
			2. '가수다'와 같은 어절이 포함된다.
			'''
			#print(morphs, morph_tags, morph, tag)
			try:
				if morph_tags[-1] == tag:  # 이미 복원된 것과 연결
					morphs.append(morphs.pop()+morph) # 미룬다 -> (미루 + ㄴ) + 다 -> 미루 + ㄴ다
				elif morph_tags[-1] == 'NN' and tag == 'EE': # '가수다'
					morphs.append('이')
					morph_tags.append('VB')
					morphs.append(morph)
					morph_tags.append(tag)
				elif morph_tags[-1] == 'MM' and tag == 'XV': # 제거 할 것, 'MMXV' 대응
					morph_tags[-1] = 'NN'
					morphs.append(morph)
					morph_tags.append(tag)
				else:
					morphs.append(morph)
					morph_tags.append(tag)
					#print('9>', i, morphs, morph_tags)
			except:
				print('>>>', morphs, morph, tag)
			return morphs, morph_tags

		def handling_co_tags(self, morphs, morph_tags, m_1, t_1, m, t, m__1, t__1):
			"""
			CO tag를 다룬다. CO tag의 형태소를 확장한다.
			"""
			#print(morphs, morph_tags, m, t)
			#------------------------------------------------------------
			def get_best_path(l):
				max_p = -1000; max_list = []; max_same_morph = 10
				for idx, x in enumerate(l):
					same_morph = 0
					_m_t_ = etc.parse_morph(x) # '가/VB+ㄴ/EE' -> [(가, VB), (ㄴ, EE)]
					#print(morphs, morph_tags, m_1, t_1, _m_t_)
					# 이전 형태소
					if (t_1 == _m_t_[0][1]) or (t_1 in ['JJ']):
						same_morph = -1 if len(morph_tags)>1 else 0
						first_word = _m_t_[0][0]
						#print(same_morph, morph_tags[same_morph])
						while morph_tags[same_morph] == _m_t_[0][1] or morph_tags[same_morph] in ['JJ']:
							first_word = morphs[same_morph] + first_word
							same_morph -= 1
							if (len(morph_tags)+same_morph)<0: break
						prev_word = (morphs[same_morph]+'/'+morph_tags[same_morph]) if (len(morph_tags)+same_morph)>=0 else 'BOS'
						prev_tag = morph_tags[same_morph] if (len(morph_tags)+same_morph)>=0 else 'BOS'
						first_word = first_word+'/'+_m_t_[0][1]
					else:
						prev_word = m_1+'/'+t_1 # viterbi를 위해서
						prev_tag = t_1
						first_word = (_m_t_[0][0]+'/'+_m_t_[0][1])

					first_tag = _m_t_[0][1]
					last_word = _m_t_[-1][0]+'/'+_m_t_[-1][1]
					last_tag = _m_t_[-1][1]
					p = (self.prob_dict[prev_word] if prev_word in self.prob_dict else -100)
					#print(p)
					p += (self.prob_dict[prev_tag + '' + first_tag] if prev_tag + '' + first_tag in self.prob_dict else -100) \
						+ (self.prob_dict[first_word] if first_word in self.prob_dict else -100) \
						+ (self.prob_dict[last_word] if last_word in self.prob_dict else -100) \
						+ (self.prob_dict[last_tag + '' + t__1] if last_tag + '' + t__1 in self.prob_dict else -100)
					#print(p)
					if p > max_p:
						max_p = p
						max_same_morph = same_morph
						max_list = []
						for i, (m, t) in enumerate(_m_t_):
							m = first_word.split('/')[0] if i==0 else m
							t = first_tag if i==0 else t
							max_list.append((m,t))
						#max_list = _m_t_
				#print(max_same_morph, max_list)
				return max_list, max_same_morph+1

			# ---------------------------------------------------------
			try:
				l = self.co_lexicon[m].split('|')
			except:
				morphs.append(m)
				morph_tags.append('NN')
				return morphs, morph_tags

			if len(l) == 1: # 후보가 하나일 경우
				_m_t_ = etc.parse_morph(l[0])
				for _m_, _t_ in _m_t_:
					if len(morph_tags)>1 and morph_tags[-1] == _t_:
						morphs.append(morphs.pop()+_m_)
					else:
						morphs.append(_m_)
						morph_tags.append(_t_)
				return morphs, morph_tags

			# ------- 사전에 후보가 2개 이상일 경우 ----------
			max_list, overlap_idx = get_best_path(l)

			co_morphs = [m for (m,t) in max_list]
			co_morph_tags = [t for (m,t) in max_list]
			#print(':::', overlap_idx, morphs, co_morphs)
			if overlap_idx < 0:
				morphs = morphs[:overlap_idx] + co_morphs
				morph_tags = morph_tags[:overlap_idx] + co_morph_tags
			else:
				morphs = morphs + co_morphs
				morph_tags = morph_tags + co_morph_tags

			return morphs, morph_tags

		def get_eojeol_tokens(self, tokens, tags, mode="eojeol"):
			"""
			# 복원 후 떨어진 형태소 연결, 구문분석에서 XV 형태소 연결하기
			# 사랑+하 -> 사랑하 (구문분석)

			param tokens : 음절
			param tags   : 품사
			"""
			eojeols, morphs, morph_tags = self.get_morph_tokens(tokens, tags, mode)
			eojeol_features = []
			#print(morphs, morph_tags)
			head_m = ''; head_t = ''; tail_m=''; tail_t=''
			for i in range(len(morphs)):
				t = morph_tags[i]
				#print(i, morphs[i], len(morphs), t)
				#if ((i==0 or morphs[i-1] == ' ') and (i == len(morphs)-1 or morphs[i+1] == ' ')):
				#	# 기호로만 한 어절이 되는 경우
				#	head_m = morphs[i]
				#	head_t = morph_tags[i]
				#	tail_m = morphs[i]
				#	tail_t = morph_tags[i]
				#	eojeol_features.append((head_m, head_t, tail_m, tail_t))
				#	print("1::::", head_m, head_t, tail_m, tail_t)
				#	continue

				# 어절 마지막
				if t == 'SP' :
					## tail feature of last eojeol
					tail_m = morphs[i-2] if (morph_tags[i-1] == 'SY' and morphs[i-1]!=',' and morphs[i-2]!= ' ') else morphs[i-1]
					tail_t = morph_tags[i-2] if (morph_tags[i-1] == 'SY' and morphs[i-1]!=',' and morphs[i-2] != ' ') else morph_tags[i-1]
					tail_t = 'EE' if tail_t in ['XV', 'VB'] else tail_t		# 비상 종결인 경우
					eojeol_features.append((head_m, head_t, tail_m, tail_t))
					#print("2::::", head_m, head_t, tail_m, tail_t)
					continue

				if i == len(morphs)-1:
					## tail feature of last eojeol
					tail_m = morphs[i-1] if (morph_tags[i] == 'SY' and morphs[i]!=',' and morphs[i-1]!= ' ') else morphs[i]
					tail_t = morph_tags[i-1] if (morph_tags[i] == 'SY' and morphs[i]!=',' and morphs[i-1]!= ' ') else morph_tags[i]
					tail_t = 'EE' if tail_t in ['XV', 'VB'] else tail_t		# 비상 종결인 경우
					eojeol_features.append((head_m, head_t, tail_m, tail_t))
					#print("3::::", head_m, head_t, tail_m, tail_t)
					continue

				# 어절 처음
				if i == 0 or morphs[i-1] == ' ':
					head_m = morphs[i+1] if (morph_tags[i] == 'SY' and morph_tags[i+1] != 'SP') else morphs[i]
					head_t = morph_tags[i+1] if (morph_tags[i] == 'SY' and morph_tags[i+1] != 'SP') else morph_tags[i]
					idx = 2 if ((morph_tags[i] == 'SY' and morph_tags[i+1] != 'SP') and i < len(morphs)-2) else 1
					head_t += morph_tags[i+idx] if morph_tags[i+idx] in ['XV', 'VB'] else ''
					#print("4:::", i, idx, morph_tags, head_m, head_t, tail_m, tail_t)

			#print(eojeols, eojeol_features)
			return eojeols, eojeol_features



"""
string2string code
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

"""
    This module contains the hash functions used in search algorithms.

    A hash function takes a string (or other object) and returns a number. 
    The number is called the hash value, hash code, or simply the hash. The hash value is used to determine the location of the string in the hash table.
    - The hash function must be deterministic, meaning that the same string always produces the same hash value. 
    - If two strings produce the same hash value, we say that the hash values collide. 
    - The hash function must also be fast, so it is important to keep the number of operations to a minimum.
"""

from typing import List, Union, Tuple, Optional
import numpy as np


# A parent class for all hash functions
class HashFunction:
    """
    This class contains the parent class for all hash functions.
    """
    def __init__(self):
        pass

    def compute(self,
        str1: str,
    ) -> int:
        """
        Returns the hash value of a string.

        Arguments:
            str1 (str): The string.

        Returns:
            int: The hash value of the string.
        """
        pass


# Polynomial rolling hash function class
class PolynomialRollingHash(HashFunction):
    """
    This class contains the polynomial rolling hash function.
    """

    def __init__(self,
        base: int = 10, # 256,
        modulus: int = 101, # 65537,
    ) -> None:
        """
        Initializes the polynomial rolling hash function.

        Arguments:
            base (int): The base to use. Default is 256.
            modulus (int): The modulus to use. Default is 65537.

        Returns:
            None

        .. note::
            * Why 65537? Because it is a Fermat prime.
        """
        super().__init__()

        # Check the inputs
        assert base > 0, 'The base must be positive.'
        assert modulus > 0, 'The modulus must be positive.'
        
        # Set the attributes
        self.base = base
        self.modulus = modulus

        # Initialize the current hash value
        self.current_hash = 0


    def compute(self,
        str1: str,
    ) -> int:
        """
        Returns the hash value of a string.

        Arguments:
            str1 (str): The string.

        Returns:
            int: The hash value of the string.
        """
        # Compute the hash value of the string
        for char in str1:
            self.current_hash = (self.current_hash * self.base + ord(char)) % self.modulus

        # Return the hash value
        return self.current_hash
    

    def update(self,
        old_char: str,
        new_char: str,
        window_size: int,
    ) -> int:
        """
        Updates the hash value of a string.

        Arguments:
            old_char (str): The old character.
            new_char (str): The new character.

        Returns:
            int: The hash value of the string.
        """
        # Update the hash value of the string
        self.current_hash = (self.current_hash - ord(old_char) * (self.base ** (window_size - 1))) % self.modulus
        self.current_hash = (self.current_hash * self.base + ord(new_char)) % self.modulus

        # Return the hash value
        return self.current_hash
    

    def reset(self) -> None:
        """
        Resets the hash value.

        Arguments:
            None

        Returns:
            None
        """
        # Reset the current hash value
        self.current_hash = 0


import unicodedata

CHOSEONG_IDX_CODEMAP = [1, 2, 0, 3, 0, 0, 4, 5, 6, 0, 0, 0, 0, 0, 0, 0, 7, 8, 9, 0, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
JONGSEONG_IDX_CODEMAP= [1, 2, 3, 4, 5, 6, 7, 0, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 0, 18, 19, 20, 21, 22, 0, 23, 24, 25, 26, 27]
getCJamoIdxChoseong  = lambda x: ((x > min(CHOSEONG_IDX_CODEMAP) and x <= max(CHOSEONG_IDX_CODEMAP)) and CHOSEONG_IDX_CODEMAP.index(x)) or 0
getCJamoIdxJongseong = lambda x: ((x > min(JONGSEONG_IDX_CODEMAP) and x <= max(JONGSEONG_IDX_CODEMAP)) and JONGSEONG_IDX_CODEMAP.index(x)) or 0



def error():

	try:
		raise Exception("function expect a character, check the value")
	except Exception as e:
		print(e)
		return

#한글 문자 판별 함수. 구: kor_check()
def is_kor_char(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	ch = ord(character)

	return (	( ch >= 0xac00 and ch <= 0xd7a3)	or	# Hangul Syllables
				( ch >= 0x1100 and ch <= 0x11ff)	or	# Hangul Jamo
				( ch >= 0x3131 and ch <= 0x318e)	or	# Hangul Compatibility Jamo
				( ch >= 0xffa1 and ch <= 0xffdc)	)	# Hangul Halfwidth


# 초/중/종성 분할 함수. kor_split()
def split_syllable(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	char = character
	returnChr = lambda x: (x and chr(x)) or str()
	returnCJJ = lambda x, y, z: tuple(map(returnChr, (x, y, z)))

	ch = ord(char)


	#초성 : comp
	if (ch >= 0x3131 and ch <= 0x314e) or (ch >= 0x3165 and ch <= 0x3186):
		return returnCJJ(ch, 0, 0)

	#중성 : comp
	if (ch >= 0x314f and ch <= 0x3163) or (ch >= 0x3187 and ch <= 0x318e):
		return returnCJJ(0, ch, 0)

	# Hangul Syllables : 가 - 힣
	if (ch >= 0xac00 and ch <= 0xd7a3):
		idx_cho = int((ch - 0xac00) / 0x024c) # idx_cho = int (ch-44032)/588
		idx_jung= int(((ch - 0xac00) % 0x024c) / 0x001c)#idx_jung = int ((ch-44032)%588)/28
		idx_jong= int((ch - 0xac00) % 0x001c)# idx_jong = int ((ch-44032) % 28)
		return returnCJJ(getCJamoIdxChoseong(idx_cho+1)+0x3131, idx_jung+0x314f, (idx_jong and getCJamoIdxJongseong(idx_jong)+0x3131) or 0)

	# None
	return returnCJJ(ch, 0, 0)



#초/중/종성 결합 함수. kor_join()
def join_syllable(choseong, jungseong, jongseong, encoding = None):


	if len(choseong)|len(choseong)|len(choseong)>1 :
		error()
		return
	elif len(choseong)|len(choseong)|len(choseong)<=0 :
		error()
		return

	returnChr = lambda x: (x and chr(x)) or str()
	returnChar = lambda x: returnChr(x)

	if not jungseong:
		if not choseong:
			return returnChar(0)
		return choseong
	else:
		if not choseong:
			return jungseong

	idx_cho  = CHOSEONG_IDX_CODEMAP[ord(choseong)-0x3131]-1
	idx_jung = ord(jungseong)-0x314f
	idx_jong = (jongseong and JONGSEONG_IDX_CODEMAP[ord(jongseong)-0x3131]) or 0

	return returnChar(0xac00+((idx_cho*21)+idx_jung)*28+idx_jong)



#한글 비교 함수
def kor_cmp(s1, s2, encoding = None):

	if len(s1)|len(s2)>1 or len(s1)|len(s2)<=0 :
		error()
		return


	if type(s1) == str:
		s1 = str().join(map(lambda x: str().join(map(lambda y: y or " ", split_syllable(x))), s1))
	if type(s2) == str:
		s2 = str().join(map(lambda x: str().join(map(lambda y: y or " ", split_syllable(x))), s2))

	return (s1>s2)-(s1<s2)



#한글 음절 판별 함수
def is_kor_syllable(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return "HANGUL SYLLABLE" in unicodedata.name(character)



#한자 문자 판별 함수
def is_hanja(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return "CJK" in unicodedata.name(character)



#숫자 판별 함수
def is_number(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return "DIGIT" in unicodedata.name(character)


#영어 알파벳 문자 판별 함수
def is_eng_char(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return "LATIN" in unicodedata.name(character)


#기호 판별 함수
def is_symbol(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return unicodedata.category(character)[0] == "S"


#구두점 판별 함수
def is_punctuation(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return unicodedata.category(character)[0] == "P"


#영어 알파벳 연결 문자 판별 함수
def is_engConnection(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return character in (".", "-", "_", "|")


# 숫자 연결 문자 판별 함수
def is_numConnection(character, encoding = None):

	if len(character)>1 or len(character)<=0:
		error()
		return

	return character in (".", ",")



#-*-coding:utf8-*-
# Filename: cider.py
#
# Description: Describes the class to compute the CIDEr (Consensus-Based Image Description Evaluation) Metric
#               by Vedantam, Zitnick, and Parikh (http://arxiv.org/abs/1411.5726)
#
# Creation Date: Sun Feb  8 14:16:54 2015
#
# Authors: Ramakrishna Vedantam <vrama91@vt.edu> and Tsung-Yi Lin <tl483@cornell.edu>

from nltkor.cider.cider_scorer import CiderScorer

class Cider:
    """
    Main Class to compute the CIDEr metric

    """
    def __init__(self, test=None, refs=None, n=4, sigma=6.0):
        # set cider to sum over 1 to 4-grams
        self._n = n
        # set the standard deviation parameter for gaussian penalty
        self._sigma = sigma

    def compute_score(self, gts, res):
        """
        Main function to compute CIDEr score
        :param  hypo_for_image (dict) : dictionary with key <image> and value <tokenized hypothesis / candidate sentence>
                ref_for_image (dict)  : dictionary with key <image> and value <tokenized reference sentence>
        :return: cider (float) : computed CIDEr score for the corpus
        """
        '''
        gts={0:['안 녕 하 세 요 홍성태입니다.','안 녕 해 요 홍성태입니다']}
        res={0:['a b ']}
        '''
        assert(gts.keys() == res.keys())
        imgIds = gts.keys()

        cider_scorer = CiderScorer(n=self._n, sigma=self._sigma)

        for id in imgIds:

            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            assert(len(ref) > 0)

            cider_scorer += (hypo[0], ref)

        score = cider_scorer.compute_score()

        return format(score,".7f")



"""
string2string similarity
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

"""
This module contains the classes for the similarity metrics and functions.
"""


from typing import List, Union, Tuple, Optional
import numpy as np

# # Import the LongestCommonSubsequence class

# for dev purposes
import sys
# sys.path.append("/Users/dowon/nltk_ko/nltk/metrics")
from nltkor.alignment import LongestCommonSubsequence, LongestCommonSubstring
# from alignment import LongestCommonSubsequence, LongestCommonSubstring

# Longest Common Subsequence based similarity class
class LCSubsequenceSimilarity(LongestCommonSubsequence):
    """
    This class contains the Longest Common Subsequence similarity metric.

    This class inherits from the LongestCommonSubsequence class.
    """

    def __init__(self):
        super().__init__()


    def compute(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        denominator: str = 'max',
    ) -> float:
        """
        Returns the LCS-similarity between two strings.

        Arguments:
            str1 (Union[str, List[str]]): The first string or list of strings.
            str2 (Union[str, List[str]]): The second string or list of strings.
            denominator (str): The denominator to use. Options are 'max' and 'sum'. Default is 'max'.

        Returns:
            float: The similarity between the two strings.

        Raises:
            ValueError: If the denominator is invalid.
        """

        # Get the numerator
        numerator, _ = super().compute(str1, str2)

        if denominator == 'max':
            return (numerator / max(len(str1), len(str2)))
        elif denominator == 'sum':
            return (2. * numerator / (len(str1) + len(str2)))
        else:
            raise ValueError('Invalid denominator.')



# Longest Common Substring based similarity class
class LCSubstringSimilarity(LongestCommonSubstring):
    """
    This class contains the Longest Common Substring similarity metric.

    This class inherits from the LongestCommonSubstring class.
    """
    def __init__(self):
        super().__init__()


    def compute(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        denominator: str = 'max',
    ) -> float:
        """
        Returns the LCS-similarity between two strings.

        Arguments:
            str1 (Union[str, List[str]]): The first string or list of strings.
            str2 (Union[str, List[str]]): The second string or list of strings.
            denominator (str): The denominator to use. Options are 'max' and 'sum'. Default is 'max'.

        Returns:
            float: The similarity between the two strings.

        Raises:
            ValueError: If the denominator is invalid.
        """
        # Get the numerator
        numerator, _ = super().compute(str1, str2)

        if denominator == 'max':
            return (numerator / max(len(str1), len(str2)))
        elif denominator == 'sum':
            return (2. * numerator / (len(str1) + len(str2)))
        else:
            raise ValueError('Invalid denominator.')


# Jaro similarity class
class JaroSimilarity:
    """
    This class contains the Jaro similarity metric.
    """

    def __init__(self):
        pass


    def compute(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> float:
        """
        This function returns the Jaro similarity between two strings.

        Arguments:
            str1 (Union[str, List[str]]): The first string or list of strings.
            str2 (Union[str, List[str]]): The second string or list of strings.

        Returns:
            float: The Jaro similarity between the two strings.
        """
        # Get the length of the strings
        len1 = len(str1)
        len2 = len(str2)

        # Get the maximum distance, which we denote by k
        k = max(len1, len2) // 2 - 1

        # Initialize the number of matching characters and the number of transpositions
        num_matches = 0
        num_transpositions = 0

        # Initialize the list of matching flags for the strings
        matches1 = [False] * len1
        matches2 = [False] * len2

        # Loop through the characters in the first string and find the matching characters
        for i in range(len1):
            # Get the lower and upper bounds for the search
            lower_bound = max(0, i - k)
            upper_bound = min(len2, i + k + 1)

            # Loop through the characters in the second string
            for j in range(lower_bound, upper_bound):
                # Check if the characters match
                if not matches2[j] and str1[i] == str2[j]:
                    # Increment the number of matches
                    num_matches += 1

                    # Set the matching flags
                    matches1[i] = True
                    matches2[j] = True

                    # Break out of the loop
                    break

        # Check if there are no matches
        if num_matches == 0:
            return 0.

        # Loop through again but this time find the number of transpositions
        # That is, the number of times where there are two matching characters but there is another "matched" character in between them
        moving_index = 0
        for i in range(len1):
            # Check if the character is a match
            if matches1[i]:
                # Find the next match
                for j in range(moving_index, len2):
                    # Check if the character is a match
                    if matches2[j]:
                        # Set the moving index
                        moving_index = j + 1

                        # Check if the characters are not in the right order
                        if str1[i] != str2[j]:
                            # Increment the number of transpositions
                            num_transpositions += 1

                        # Break out of the loop
                        break

        num_transpositions = num_transpositions // 2

        # Return the Jaro similarity
        return (num_matches / len1 + num_matches / len2 + (num_matches - num_transpositions) / num_matches) / 3.0

def demo():
    """
    This function demonstrates the similarity metrics.
    """
    # Initialize the similarity metrics
    lcs_sim = LCSubsequenceSimilarity()
    lcs_sub_sim = LCSubstringSimilarity()
    jaro_sim = JaroSimilarity()

    # Initialize the strings
    str1 = '제가 나와 있는 곳은 경남 거제시 옥포동 덕포 해수욕장에 나와 있습니다.'
    str2 = '강한 바람에 간판이나 지붕이 떨어지는 등 피해가 잇따르기도 했습니다.'

    # Get the similarity metrics
    lcs_sim_score = lcs_sim.compute(str1, str2)
    lcs_sub_sim_score = lcs_sub_sim.compute(str1, str2)
    jaro_sim_score = jaro_sim.compute(str1, str2)

    # Print the results
    print('Longest Common Subsequence Similarity: {}'.format(lcs_sim_score))
    print('Longest Common Substring Similarity: {}'.format(lcs_sub_sim_score))
    print('Jaro Similarity: {}'.format(jaro_sim_score))

if __name__ == '__main__':
    demo()



import re



def parse_morph(target):
		
	result=list()
	buf = target
	buf = re.sub(' ','',buf)
	loop = re.search("(\/([A-Za-z]+)(?:\+|\n|$))", buf)

	while loop:
		pos = buf.find(loop.group(1))
		spos = buf[:pos].rfind("\t") + 1
		result.append((buf[spos:pos], loop.group(2)))
		buf = buf[pos+len(loop.group(1)):]
		loop = re.search("(\/([A-Za-z]+)(?:\+|\n|$))",buf)
	return result







import datetime
import os

def make_requirement(packages):
    file_path = os.path.abspath(__file__)
    file_path = file_path + "__requirement__NLTKor.txt"
    with open(file_path, "a") as f:
        for package in packages:
            f.write(package + '\n')

    return file_path


from nltkor import alignment
from nltkor import cider
from nltkor import distance
from nltkor import sejong
from nltkor import metrics
from nltkor import misc
from nltkor import search
from nltkor import similarity
from nltkor import tag
from nltkor import tokenize


from nltkor import trans
from nltkor import Kor_char
from nltkor import etc



from nltkor.tag import EspressoTagger

class EntMent :
    
    def __init__(self):
        self.entity_list = []
        self.tagger = EspressoTagger(task = 'ner')
        self.skip = ["*"]
    
    def entity (self,orginal_text,summarized_text) :
    
        self.entity_list = self.tagger.tag(orginal_text)
        self.entity_list = [item for item in self.entity_list if item[1] not in self.skip]
        self.entity_recall(summarized_text)
        
    def entity_recall (self,summarized_text) :
        
        if len(self.entity_list) == 0 :
            return 0.0
        
        summarized_entity_list = self.tagger.tag(summarized_text)
        summarized_entity_list = [item for item in summarized_entity_list  if item[1] not in self.skip]
        recall = len(set(summarized_entity_list)&set(self.entity_list))/len(summarized_entity_list)
        print(recall)



# The following trick allows us to import the classes directly from the search module:
from .classical import (
    SearchAlgorithm,
    NaiveSearch,
    RabinKarpSearch,
    KMPSearch,
    BoyerMooreSearch,
)
from .faiss_search import FaissSearch
from .kobert_tokenizer import KoBERTTokenizer


import zipfile
import base64
import os
import requests

class SejongDir :
    
        def __init__(self, data_dir=None):
                    self.data_dir = data_dir
                    if data_dir == None:
                            path= os.path.join(os.path.dirname(__file__), 'sejong_dictionary')
                            self.data_dir = path
                        
                    self.path = ""

                    if not self._check_sejong_dictionary():
                        self._download_sejong_dictionary()

                    
        def _check_sejong_dictionary(self):
            """Checks if the sejong dictionary is available and downloads it if necessary"""
            if not os.path.exists(self.data_dir+"/sejong_dictionary.zip"):
                    return False
            else:
                    return True
                    
                    
        def _download_sejong_dictionary(self):
            """Downloads the sejong xml from the server"""
            temp_path = os.path.join(os.path.dirname(__file__),'sejong_dictionary.zip')
            url = "https://air.changwon.ac.kr/~airdemo/storage/sejong/sejong_dictionary.zip"
            print("Downloading sejong dictionary...")
            with requests.get(url, stream=True) as r:
                    r.raise_for_status()
                    with open(temp_path, 'wb') as f:
                            for chunk in r.iter_content(chunk_size=8192):
                            # If you have chunk encoded response uncomment if
                            # and set chunk_size parameter to None.
                            #if chunk:
                                    f.write(chunk)
            self.unzip_with_correct_encoding(temp_path)
            
            
        def unzip_with_correct_encoding(self, zip_path):
            """Unzips a ZIP file and decodes filenames with proper encoding"""
            extract_to_dir = self.data_dir
 
            if not os.path.exists(extract_to_dir):
                os.makedirs(extract_to_dir)

            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                for file in zip_ref.namelist():
                    
                    if file.endswith('/'):

                        try:
                            decoded_dir = file.encode('cp437').decode('utf-8')
                        except Exception as e:
                            print(f"디렉토리명 디코딩 실패: {file} - {e}")
                            decoded_dir = file 

                        dir_path = os.path.join(extract_to_dir, "")
                        os.makedirs(dir_path, exist_ok=True)
                        continue
                    
                    try:
                        decoded_name = file.encode('cp437').decode('utf-8')
                    except Exception as e:
                        print(f"파일명 디코딩 실패: {file} - {e}")
                        decoded_name = file  

                    # 추출 경로 설정
                    extracted_path = os.path.join(extract_to_dir, decoded_name)
                    extracted_path = extracted_path.replace("sejong_xml_/", "")
                    # 파일 저장
                    with zip_ref.open(file) as original_file:
                        os.makedirs(os.path.dirname(extracted_path), exist_ok=True)
                        with open(extracted_path, "wb") as f:
                            f.write(original_file.read())
            

                    

                    






# import to provide easier access for nlpnet users
from .config import set_data_dir
from . import taggers
from . import utils

from .taggers import POSTagger, NERTagger, WSDTagger, SRLTagger, DependencyParser
from .utils import tokenize

__version__ = '1.2.0'


"""
string2string similarity
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

from typing import List, Union, Tuple
import torch
from torch import Tensor
from torch.nn import functional as F
import numpy as np

# for dev purposes
import sys
# sys.path.append("/Users/dowon/nltk_ko/nltk/misc")
from nltkor.misc.string2string_word_embeddings import GloVeEmbeddings
# from string2string_word_embeddings import GloVeEmbeddings


# Cosine similarity class
class CosineSimilarity:
    def __init__(self) -> None:
        r"""
        This function initializes the CosineSimilarity class.
        """
        pass


    # Compute (tensor)
    def _compute_tensor(self,
        x1: Tensor,
        x2: Tensor,
        dim: int = 1,
        eps: float = 1e-8
    ) -> Tensor:
        r"""
        Computes the cosine similarity between two tensors along a given dimension.

        Arguments:
            x1 (Tensor): First tensor.
            x2 (Tensor): Second tensor.
            dim (int): Dimension to compute cosine similarity.
            eps (float): Epsilon value.

        Returns:
            Tensor: Cosine similarity between two tensors along a given dimension.
        """
        # Make sure that x1 and x2 are float tensors
        if x1.dtype != torch.float:
            x1 = x1.float()
        if x2.dtype != torch.float:
            x2 = x2.float()
        # Compute cosine similarity between two tensors
        return F.cosine_similarity(x1, x2, dim, eps)


    # Compute (numpy)
    def _compute_numpy(self,
        x1: np.ndarray,
        x2: np.ndarray,
        dim: int = 1,
        eps: float = 1e-8
    ) -> np.ndarray:
        r"""
        Computes the cosine similarity between two numpy arrays along a given dimension.

        Arguments:
            x1 (np.ndarray): First numpy array.
            x2 (np.ndarray): Second numpy array.
            dim (int): Dimension (or axis in the numpy realm) to compute cosine similarity.
            eps (float): Epsilon value (to prevent division by zero).

        Returns:
            np.ndarray: Cosine similarity between two numpy arrays along a given dimension.
        """
        # Compute cosine similarity between two numpy arrays along a given dimension "dim"
        return np.sum(x1 * x2, axis=dim) / np.maximum(np.linalg.norm(x1, axis=dim) * np.linalg.norm(x2, axis=dim), eps)


    # Compute
    def compute(self,
        x1: Union[Tensor, np.ndarray],
        x2: Union[Tensor, np.ndarray],
        dim: int = 0,
        eps: float = 1e-8
    ) -> Union[Tensor, np.ndarray]:
        r"""
        Computes the cosine similarity between two tensors (or numpy arrays) along a given dimension.

        * For two (non-zero) vectors, :math:`x_1` and :math:`x_2`, the cosine similarity is defined as follows:

            .. math::
                :nowrap:

                \begin{align}
                    \texttt{cosine-similarity}(x_1, x_2) & = |x_1|| \ ||x_2|| \cos(\theta) \\
                    & = \frac{x_1 \cdot x_2}{||x_1|| \ ||x_2||} \\
                    & = \frac{\sum_{i=1}^n x_{1i} x_{2i}}{\sqrt{\sum_{i=1}^n x_{1i}^2} \sqrt{\sum_{i=1}^n x_{2i}^2}}
                \end{align}

            where :math:`\theta` denotes the angle between the vectors, :math:`\cdot` the dot product, and :math:`||\cdot||` the norm operator.

        * In practice, the cosine similarity is computed as follows:

            .. math::
                :nowrap:

                \begin{align}
                    \texttt{cosine-similarity}(x_1, x_2) & = \frac{x_1 \cdot x_2}{\max(||x_1|| ||x_2||, \epsilon)}
                \end{align}

            where :math:`\epsilon` is a small value to avoid division by zero.


        Arguments:
            x1 (Union[Tensor, np.ndarray]): First tensor (or numpy array).
            x2 (Union[Tensor, np.ndarray]): Second tensor (or numpy array).
            dim (int): Dimension to compute cosine similarity (default: 0).
            eps (float): Epsilon value (to avoid division by zero).

        Returns:
            Union[Tensor, np.ndarray]: Cosine similarity between two tensors (or numpy arrays) along a given dimension.

        Raises:
            TypeError: If x1 and x2 are not of the same type (either tensor or numpy array).
            TypeError: If x1 and x2 are not tensors or numpy arrays.
        """
        # Check if x1 and x2 are of the same type (either tensor or numpy array)
        if type(x1) != type(x2):
            raise TypeError("x1 and x2 must be of the same type (either tensor or numpy array).")

        # If x1 and x2 are tensors
        if type(x1) == Tensor:
            # Compute cosine similarity
            return self._compute_tensor(x1, x2, dim, eps)
        # If x1 and x2 are numpy arrays
        elif type(x1) == np.ndarray:
            # Compute cosine similarity
            return self._compute_numpy(x1, x2, dim, eps)
        # If x1 and x2 are not tensors or numpy arrays
        else:
            raise TypeError("x1 and x2 must be either tensors or numpy arrays.")

def demo():
    array1 = np.array([20, 65, 1])
    array2 = np.array([98, 67, 548])

    print("demo : ", CosineSimilarity().compute(array1, array2))

if __name__ == "__main__":
    demo()



# -*- coding: utf-8 -*-

"""
This script contains the definition of the Metadata class.
It can also be invoked in order to create a Metada object
and save it to a file in the data directory.
"""

import _pickle

from . import config

class Metadata(object):
    """
    Class for storing metadata about a neural network and its 
    parameter files.
    """
    
    def __init__(self, task, paths=None, use_caps=True, use_suffix=False, use_prefix=False,
                 use_pos=False, use_chunk=False, use_lemma=False):
        self.task = task
        self.paths = paths if paths else config.FILES
        self.use_caps = use_caps
        self.use_suffix = use_suffix
        self.use_prefix = use_prefix
        self.use_pos = use_pos
        self.use_chunk = use_chunk
        self.use_lemma = use_lemma
        self.metadata = 'metadata_%s' % task
        self.network = 'network_%s' % task
        self.network_text = 'network_text_%s' % task
        self.tag_dict = '%s_tag_dict' % task
        
        # dependency edge filter doesn't use an actual neural network, so 
        # we call it "model" to be more consistent
        self.model = self.network
        '''
        if task == 'srl_boundary':
            self.pred_dist_table = 'pred_dist_table_boundary'
            self.target_dist_table = 'target_dist_table_boundary'
            self.transitions = 'srl_transitions_boundary'
            self.type_features = 'type_features_boundary'
            self.caps_features = 'caps_features_boundary'
            self.pos_features = 'pos_features_boundary'
            self.chunk_features = 'chunk_features_boundary'
            self.suffix_features = None
            
        elif task == 'srl_classify':
            self.pred_dist_table = 'pred_dist_table_classify'
            self.target_dist_table = 'target_dist_table_classify'
            self.transitions = None
            self.type_features = 'type_features_classify'
            self.caps_features = 'caps_features_classify'
            self.pos_features = 'pos_features_classify'
            self.chunk_features = 'chunk_features_classify'
            self.suffix_features = None
        
        elif task == 'srl':
            # one step srl
            self.pred_dist_table = 'pred_dist_table_1step'
            self.target_dist_table = 'target_dist_table_1step'
            self.transitions = 'srl_transitions_1step'
            self.type_features = 'type_features_1step'
            self.caps_features = 'caps_features_1step'
            self.pos_features = 'pos_features_1step'
            self.chunk_features = 'chunk_features_1step'
            self.suffix_features = None

        else:
            self.type_features = 'type_features_%s' % task
            self.caps_features = 'caps_features_%s' % task
            self.pos_features = 'pos_features_%s' % task
            self.chunk_features = 'chunk_features_%s' % task
            self.suffix_features = 'suffix_features_%s' % task
            self.prefix_features = 'prefix_features_%s' % task
        '''
        self.type_features = 'type_features_%s' % task
        self.caps_features = 'caps_features_%s' % task
        self.pos_features = 'pos_features_%s' % task
        #self.chunk_features = 'chunk_features_%s' % task
        self.suffix_features = 'suffix_features_%s' % task
        self.prefix_features = 'prefix_features_%s' % task

    def __str__(self):
        """Shows the task at hand and which attributes are used."""
        lines = []
        lines.append("Metadata for task %s" % self.task)
        for k in self.__dict__:
            if isinstance(k, str) and k.startswith('use_'):
                lines.append('%s: %s' % (k, self.__dict__[k]))
        
        return '\n'.join(lines)
    
    def save_to_file(self): 
        """
        Save the contents of the metadata to a file. The filename is determined according
        to the task.
        """
        save_data = self.__dict__.copy()
        filename = self.paths['metadata_%s' % self.task]
        del(save_data['paths'])
        
        with open(filename, 'wb') as f:
            _pickle.dump(save_data, f, 2)

        filename = self.paths['metadata_text_%s' % self.task]
        
        with open(filename, 'wt') as f:
            for k,v in save_data.items():
                f.write("%s: %s\n" %(k,v))
    
    @classmethod
    def load_from_file(cls, task, paths=None):
        """
        Reads the file containing the metadata for the given task and returns a 
        Metadata object.
        """
        if paths is None:
            paths = config.FILES
        md = Metadata(None, paths)
        
        # the actual content of the file is the __dict__ member variable, which contain all
        # the instance's data
        with open(paths['metadata_%s' % task], 'rb') as f:
            data = _pickle.load(f)
        md.__dict__.update(data)
        
        return md




import sys
from setuptools import Extension
from setuptools import setup
from setuptools import find_packages
from setuptools import dist
import os
dist.Distribution().fetch_build_eggs(['numpy>=1.10'])
import numpy as np

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('nltkor/sejong')
extra_files = extra_files+package_files('nltk_kor/tag')

module1 = Extension("nltkor.tag.libs.network",
                               ["nltkor/tag/libs/network.c"],
                               include_dirs=['.', np.get_include()])

setup(
  name='nltkor',
  version='1.2.4',
	url='https://modi.changwon.ac.kr/air_cwnu/nlp_tool/nltk_ko.git',
  packages=find_packages(exclude=[]),
  python_requires='>=3.7',
  install_requires=[
    'regex',
    'tqdm>=4.40.0',
    'joblib',
    'numpy==1.23.0',
    'requests',
    'nltk > 3.0',
    'pyarrow ==14.0.0',
    'beautifulSoup4',
    'faiss-cpu>=1.7.3',
    'datasets',
    'torch',
    'scikit-learn>=0.22.1',
    'transformers>=4.8.2',
    'protobuf',
    'sentencepiece',
    'pandas',
    'bert_score',
    'fasttext==0.9.2'
    ],
  package_data={'': extra_files},
	ext_modules=[module1],
  include_package_data=True,

  tests_require=["pytest"],
  classifiers=[
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Typing :: Typed",
  ],
  keywords=[
      "string matching",
      "pattern matching",
      "edit distance",
      "string to string correction",
      "string to string matching",
      "Levenshtein edit distance",
      "Hamming distance",
      "Damerau-Levenshtein distance",
      "Jaro-Winkler distance",
      "longest common subsequence",
      "longest common substring",
      "dynamic programming",
      "approximate string matching",
      "semantic similarity",
      "natural language processing",
      "NLP",
      "information retrieval",
      "rouge",
      "sacrebleu",
      "bertscore",
      "bartscore",
      "fasttext",
      "glove",
      "cosine similarity",
      "Smith-Waterman",
      "Needleman-Wunsch",
      "Hirschberg",
      "Karp-Rabin",
      "Knuth-Morris-Pratt",
      "Boyer-Moore",
  ],
)



# -*- coding: utf-8 -*-

'''
Class for dealing with dependency parsing data.
'''

import os
import logging
import numpy as np

from .. import attributes
from .. import reader
from ..word_dictionary import WordDictionary

class ConllPos(object):
	'''
	Dummy class to store field positions in a CoNLL-like file
	for dependency parsing. NB: The positions are different from
	those used in SRL!
	'''
	id = 0
	word = 1	# 어절 
	morph_h = 2	# 접두사를 제외한 첫 형태소
	pos_h = 3		# 접두사를 제외한 첫 형태소의 품사 
	morph_t = 4	# 마지막 형태소 
	pos_t = 5	# 마지막 형태소 품사
	dep_head = 6 # dependency head 의존정보 
	dep_rel = 7 # dependency relation 구문 태그 
	SEP = '\t'


class DependencyReader(reader.TaggerReader):
	'''
	Class to read dependency files in CoNLL X format.
	'''
	
	def __init__(self, md=None, filename=None, labeled=False):
		'''
		Constructor.
		:param md: Metadata object containing the description for this reader
		:param filename: file containing data to be read and used in training
				or tagging
		:param labeled: (ignored if md is supplied) whether it is intended 
				to be used in labeled dependency parsing. Note that if it is 
				True, another reader object will be needed for unlabeled dependency.
		'''
		if md is not None:
				self.labeled = md.task.startswith('labeled')
		else:
				self.labeled = labeled
		
		if filename is not None:
				self._read_conll(filename)
		
		if self.labeled:
				self.taskname = 'labeled_dependency'
		else:
				self.taskname = 'unlabeled_dependency'
		
		self.rare_tag = None
		self.pos_dict = None
		super(DependencyReader, self).__init__(md)
		
	@property
	def task(self):
		"""
		Abstract Base Class (ABC) attribute.
		"""
		return self.taskname
						
	
	def _read_conll(self, filename): 
		'''
		Read data from a CoNLL formatted file.
		'''
		lines = []
		self.sentences = []
		self.heads = []
		
		# this keeps track of the tokens 
		sentence = []
		
		# this has the number of each token's head, in the same order as 
		# the tokens appear
		sentence_heads = []
		if self.labeled:
				self.labels = []
				sentence_labels = []
		
		with open(filename, 'r') as f:
				for line in f:
						line = line.strip()
						#line = unicode(line, 'utf-8').strip()
						lines.append(line)
		
		for line in lines:
				if line == '':
						# empty line, last sentence is finished 
						if len(sentence) > 0:
								self.sentences.append(sentence)
								self.heads.append(np.array(sentence_heads))
								
								if self.labeled:
										self.labels.append(sentence_labels)
										sentence_labels = []
										
								sentence = []
								sentence_heads = []
								
						continue
				
				fields = line.split(ConllPos.SEP)
				word = fields[ConllPos.word]
				morph_h = fields[ConllPos.morph_h]
				pos_h = fields[ConllPos.pos_h]
				morph_t = fields[ConllPos.morph_t]
				pos_t = fields[ConllPos.pos_t]
				head = int(fields[ConllPos.dep_head])
				label = fields[ConllPos.dep_rel]
				
				if head == 0:
						# we represent a dependency to root as an edge to the token itself
						head = int(fields[ConllPos.id])
				
				# -1 because tokens are numbered from 1
				head -= 1
				
				token = attributes.Token(word, pos_h=pos_h, morph_t=morph_t, pos_t=pos_t )
				#token = attributes.Token(word, morph_h=morph_h, pos_h=pos_h, morph_t=morph_t, pos_t=pos_t )
				sentence.append(token)
				sentence_heads.append(head)
				if self.labeled:
						sentence_labels.append(label)
		
		# in case there was not an empty line after the last sentence 
		if len(sentence) > 0:
				self.sentences.append(sentence)
				self.heads.append(np.array(sentence_heads))
				if self.labeled:
						self.labels.append(sentence_labels)
	
	def _create_pos_dict(self):
		"""
		Examine all POS tags in the sentences and create a dictionary based on them.
		"""
		logger = logging.getLogger("Logger")
		logger.info('Creating new POS tag dictionary (for dependency parsing)')
		tags = {token.pos_h for sent in self.sentences
						for token in sent} | {token.pos_t for sent in self.sentences
						for token in sent}
		pos_dict = {tag: code for code, tag in enumerate(tags)}
		
		code = max(pos_dict.values()) + 1
		pos_dict[attributes.PADDING_POS] = code
		
		return pos_dict
	
	def load_pos_dict(self):
		"""
		Load the pos tag dictionary (specific for dependency parsing) 
		from its default location.
		"""
		logger = logging.getLogger("Logger")
		logger.debug('Loading POS tag dictionary (for dependency parsing)')
		pos_dict = reader.load_tag_dict(self.md.paths['dependency_pos_tags'])
		return pos_dict
	
	def load_tag_dict(self, filename=None):
		"""
		Verify if this reader is for the unlabeled dependency task. If so, 
		it doesn't use a tag dictionary and the call is ignored.
		"""
		if not self.labeled:
				return
		
		super(DependencyReader, self).load_tag_dict(filename)		
	
	def load_or_create_tag_dict(self):
		"""
		Try to load the tag dictionary from the default location. If the dictinaty
		file is not available, scan the available sentences and create a new one.
		
		It only is needed in labeled dependency parsing. 
		"""
		if not self.labeled:
				return
		
		logger = logging.getLogger('Logger')
		filename = self.md.paths['dependency_tag_dict']
		if os.path.isfile(filename):
				self.load_tag_dict(filename)
				logger.debug('Loaded dependency tag dictionary')
				return
		
		tags = {tag for sent_labels in self.labels for tag in sent_labels}
		self.tag_dict = {tag: code for code, tag in enumerate(tags)}
		
		reader.save_tag_dict(filename, self.tag_dict)
		logger.debug('Saved dependency tag dictionary')
		
	
	def generate_dictionary(self, dict_size=None, minimum_occurrences=2):
		"""
		Generates a token dictionary based on the given sentences.
		
		:param dict_size: Max number of tokens to be included in the dictionary.
		:param minimum_occurrences: Minimum number of times that a token must
				appear in the text in order to be included in the dictionary.
		"""
		logger = logging.getLogger("Logger")
		#all_tokens = [token.morph_t for sent in self.sentences for token in sent]
		all_tokens = [token.morph_h for sent in self.sentences for token in sent] \
							+ [token.morph_t for sent in self.sentences for token in sent]
		self.word_dict = WordDictionary(all_tokens, dict_size, minimum_occurrences)
		logger.info("Created dictionary with %d tokens" % self.word_dict.num_tokens)
	
	def codify_sentences(self):
		"""
		Converts each token in each sequence into indices to their feature vectors
		in feature matrices. The previous sentences as text are not accessible anymore.
		Tags are left as the index of the each token's head.
		"""
		if self.converter is None:
				self.create_converter()
		
		self.sentences = [np.array([self.converter.convert(token) for token in sent])
											for sent in self.sentences]
		
		if self.labeled:
				self.labels = [np.array([self.tag_dict[label] for label in sent_labels]) 
											 for sent_labels in self.labels]
		
		self.codified = True
	
	def _load_or_create_pos_dict(self):
		"""
		Try to load the pos tag dictionary to be used with this reader (when
		using POS tags as additional features). If there isn't a file in the 
		data directory with the right name, a new dictionary is created 
		after examining the data.
		"""
		if self.pos_dict is not None:
				return
		
		if os.path.isfile(self.md.paths['dependency_pos_tags']):
				self.pos_dict = self.load_pos_dict()
		else:
				self.pos_dict = self._create_pos_dict()
				self.save_tag_dict(self.md.paths['dependency_pos_tags'], self.pos_dict)
			
	def get_num_pos_tags(self):
		"""
		Return the number of POS tags that can be used as an additional feature
		by this reader.
		"""
		self._load_or_create_pos_dict()
		return len(self.pos_dict)
	
	def create_converter(self):
		"""
		This function overrides the TextReader's one in order to deal with Token
		objects instead of raw strings. It also allows POS as an attribute.
		여기 순서와 utils.create_feature_table의 순서가 같아야 한다. 
		"""
		#f = lambda token: self.word_dict[token.morph_h]
		self.converter = attributes.TokenConverter()
		#self.converter.add_extractor(f)

		f = lambda token: self.word_dict[token.morph_t]
		self.converter.add_extractor(f)
		
		#if self.md.use_caps:
		#		caps_lookup = lambda t: attributes.get_capitalization(t.word)
		#		self.converter.add_extractor(caps_lookup)
		
		if self.md.use_pos:
			self._load_or_create_pos_dict()
			g = lambda token: self.pos_dict[token.pos_h]
			self.converter.add_extractor(g)
			g = lambda token: self.pos_dict[token.pos_t]
			self.converter.add_extractor(g)
				
		



import os

# import sys
# sys.path.append("/Users/dowon/nltk_ko/nltk/tag")
# from libs import *

"""
This script will run a POS or SRL tagger on the input data and print the results
to stdout.
"""

import argparse
import logging

if __package__ is None: # 스크립트로 실행할 때
		import sys
		from os import path
		#print(path.dirname( path.dirname( path.abspath(__file__) ) ))
		sys.path.append(path.dirname( path.dirname( path.abspath(__file__) ) ))
		from libs import *
else:
	from .libs import *
#from .libs import *
import requests
import zipfile

logging.disable(logging.INFO) # disable INFO and DEBUG logging everywhere


class EspressoTagger:
		def __init__(self, data_dir=None, task="pos"):
				self.data_dir = data_dir
				if data_dir == None:
						path=os.path.dirname(__file__)
						path= path + '/data'
						self.data_dir = path
						#print(path)
				self.path = ""
				self.tagger = None

				self.task = task.lower()
				if not self._check_model():
						self._download_model()

				set_data_dir(self.data_dir)

				if self.task == 'pos':
								self.tagger = taggers.POSTagger(data_dir=self.data_dir)
				elif self.task == 'ner':
								self.tagger = taggers.NERTagger(data_dir=self.data_dir)
				elif self.task == 'wsd':
								self.tagger = taggers.WSDTagger(data_dir=self.data_dir)
				elif self.task == 'srl':
								self.tagger = taggers.SRLTagger(data_dir=self.data_dir)
				elif self.task == 'dependency':
								self.tagger = taggers.DependencyParser(data_dir=self.data_dir)
				else:
								raise ValueError('Unknown task: %s' % self.task)


		def tag(self, text, use_sent_tokenizer=True, lemma=True):
				"""
				This function provides an interactive environment for running the system.
				It receives text from the standard input, tokenizes it, and calls the function
				given as a parameter to produce an answer.

				:param task: 'pos', ner', 'wsd', 'srl' or 'dependency'
				:param use_tokenizer: whether to use built-in tokenizer
				"""

				#use_sent_tokenizer = not use_sent_tokenizer
				mode = 'standard' if lemma else 'eumjeol'
				result = self.tagger.tag(text, use_sent_tokenizer, mode)
				'''
				else:
								tokens = text.split()
								if self.task != 'dependency':
												result = [self.tagger.tag_tokens(tokens, True)]
								else:
												result = [self.tagger.tag_tokens(tokens)]
				'''

				return self._result_tagged(result, self.task)

		def _result_tagged(self, tagged_sents, task):
				"""
				Prints the tagged text to stdout.

				:param tagged_sents: sentences tagged according to any of espresso taggers.
				:param task: the tagging task (either 'pos', 'ner', 'wsd', 'srl' or 'dependency')
				"""

				##TODO: print부분 return으로 변경
				if task == 'pos':
								return self._return_tagged_pos(tagged_sents)
				elif task == 'ner':
								return self._return_tagged_ner(tagged_sents)
				elif task == 'wsd':
								return self._return_tagged_wsd(tagged_sents)
				elif task == 'srl':
								return self._return_tagged_srl(tagged_sents)
				elif task == 'dependency':
								return self._return_parsed_dependency(tagged_sents)
				else:
								raise ValueError('Unknown task: %s' % task)



		def _return_parsed_dependency(self, parsed_sents):
				"""Prints one token per line and its head"""
				result = []
				temp_list = []
				temp_list2 = []
				for sent in parsed_sents:
						temp_list = sent.to_conll().split('\t')
						temp_list = temp_list[1:]
						for ele in temp_list:
								if '\n' in ele:
										ele = ele[:ele.find('\n')]
								temp_list2.append(ele)
						result.append(self._dependency_after(temp_list2)[:])
						temp_list2 = []

				return result

		def _return_tagged_pos(self, tagged_sents):
			"""Prints one sentence per line as token_tag"""
			result = []
			for sent in tagged_sents:
				result = result + list(sent)
			return result

		def _return_tagged_srl(self, tagged_sents):
			result = []
			for sent in tagged_sents:
				# print (' '.join(sent.tokens))
				temp_dict1 = {}
				for predicate, arg_structure in sent.arg_structures:
					# print ("test 1 :", predicate)
					# print("te22 :", arg_structure)

					temp_dict2 = {}
					for label in arg_structure:
						argument = ' '.join(arg_structure[label])
						# line = '\t%s: %s' % (label, argument)
						# print (line)
						temp_dict2[label] = argument

						# result.append((label, argument))
						# print ('\n')
					temp_dict1[predicate] = temp_dict2

				result.append(temp_dict1)

			return result

		def _return_tagged_ner(self, tagged_sents):
				"""Prints one sentence per line as token_tag"""
				result = []
				for sent in tagged_sents:
						for item in sent:
								#s = '_'.join(item)
								result.append(item)

				return result

		def _return_tagged_wsd(self, tagged_sents):
				"""Prints one sentence per line as token_tag"""
				result = []
				for sent in tagged_sents:
						for item in sent:
								s = '_'.join(item)
								result.append(s)

				return result

		def _download_model(self):
				"""Downloads the model from the server"""
				temp_path = os.path.dirname(__file__) + '/data.zip'
				url = "https://air.changwon.ac.kr/~airdemo/storage/espresso_data_1/data.zip"
				print("Downloading Espresso5 model...")
				with requests.get(url, stream=True) as r:
						r.raise_for_status()
						with open(temp_path, 'wb') as f:
								for chunk in r.iter_content(chunk_size=8192):
								# If you have chunk encoded response uncomment if
								# and set chunk_size parameter to None.
								#if chunk:
										f.write(chunk)

				if os.path.exists(self.data_dir):
						os.rmdir(self.data_dir)

				with zipfile.ZipFile(temp_path, "r") as zip_ref:
						zip_ref.extractall(os.path.dirname(__file__))

		def _check_model(self):
				"""Checks if the model is available and downloads it if necessary"""
				if not os.path.exists(self.data_dir):
						return False
				else:
						return True

		def _dependency_after(self, list):
				len_list = len(list)
				temp_list = []
				repeat = len_list//3
				for i in range(repeat):
						index = i*3
						tup1 = (i+1, )
						tup2 = tuple(list[index:index+3])
						tup = tup1 + tup2
						temp_list.append(tup[:])


				return temp_list

if __name__ == '__main__':
		tagger = EspressoTagger(task='pos')
		print(tagger.tag("나는 아름다운 강산에 살고있다."))



# The following trick allows us to import the classes directly from the similarity module:
from .cosine_similarity import CosineSimilarity
from .classical import LCSubstringSimilarity, LCSubsequenceSimilarity, JaroSimilarity


# Author: Krishna Pillutla
# License: GPLv3
import json
import os
import time
from tqdm.auto import tqdm as tqdm_original

import torch
from transformers import AutoModel, AutoTokenizer, XLNetTokenizer
# from nltkor.search.kobert_tokenizer import KoBERTTokenizer

CPU_DEVICE = torch.device('cpu')
tqdm = lambda *args, **kwargs: tqdm_original(
    *args, **kwargs, disable=os.environ.get("DISABLE_TQDM", False))


def get_device_from_arg(device_id):
    if (device_id is not None and
            torch.cuda.is_available() and
            0 <= device_id < torch.cuda.device_count()):
        return torch.device(f'cuda:{device_id}')
    else:
        return CPU_DEVICE

def get_model(model_name, tokenizer, device_id):
    device = get_device_from_arg(device_id)
    if 'gpt2' in model_name or "bert" in model_name:
        model = AutoModel.from_pretrained(model_name, pad_token_id=tokenizer.eos_token_id).to(device)
        model = model.eval()
    else:
        raise ValueError(f'Unknown model: {model_name}')
    return model

def get_tokenizer(model_name='skt/kobert-base-v1'):
    if 'gpt2' in model_name or "bert" in model_name:
        if model_name == 'skt/kobert-base-v1':
            # tokenizer = KoBERTTokenizer.from_pretrained(model_name)
            tokenizer = XLNetTokenizer.from_pretrained(model_name)
        else:
            tokenizer = AutoTokenizer.from_pretrained(model_name)
    else:
        raise ValueError(f'Unknown model: {model_name}')
    return tokenizer

def load_json_dataset(data_path, max_num_data):
    texts = []
    for i, line in enumerate(open(data_path)):
        if i >= max_num_data:
            break
        texts.append(json.loads(line)['text'])
    return texts

def load_and_tokenize_json_data(tokenizer, data_path, max_len=1024, max_num_data=float('inf')):
    """ Load and tokenize the data in a jsonl format

    :param tokenizer:  HF tokenizer object
    :param data_path: jsonl file to read. Read the "text" field of each line
    :param max_len: maximum length of tokenized data
    :param max_num_data: maximum number of lines to load
    :return: list of `torch.LongTensor`s of shape (1, num_tokens), one for each input line
    """
    assert max_len <= 1024 and max_num_data >= 2000, f"max_len={max_len}, max_num_data={max_num_data} are insufficent"
    t1 = time.time()
    texts = load_json_dataset(data_path, max_num_data=max_num_data)
    t2 = time.time()
    print(f'dataset load time: {round(t2-t1, 2)} sec')
    t1 = time.time()
    tokenized_texts = [tokenizer.encode(sen, return_tensors='pt', truncation=True, max_length=max_len)
                      for sen in texts]
    t2 = time.time()
    print(f'tokenizing time: {round(t2-t1, 2)} sec')
    return tokenized_texts

def decode_samples_from_lst(tokenizer, tokenized_texts):
    """ Decode from tokens to string

    :param tokenizer: HF tokenizer
    :param tokenized_texts: list of list of tokens
    :return: decoded output as a list of strings of the same length as tokenized_text_list
    """
    t1 = time.time()
    output = []
    for l in tokenized_texts:
        o = tokenizer.decode(torch.LongTensor(l), skip_special_tokens=True)
        output.append(o)
    t2 = time.time()
    print(f'de-tokenizing time: {round(t2-t1, 2)}')
    return output

@torch.no_grad()
def featurize_tokens_from_model(model, tokenized_texts, batch_size, name="", verbose=False):
    """Featurize tokenized texts using models, support batchify
    :param model: HF Transformers model
    :param batch_size: Batch size used during forward pass
    :param tokenized_texts: list of torch.LongTensor of shape (1, length)
    :param verbose: If True, print status and time
    :return:
    """
    device = next(model.parameters()).device
    t1 = time.time()
    feats, chunks, chunk_sent_lengths = [], [], []
    chunk_idx = 0

    while chunk_idx * batch_size < len(tokenized_texts):
        _chunk = [_t.view(-1) for _t in tokenized_texts[chunk_idx * batch_size: (chunk_idx + 1) * batch_size]]
        chunks.append(_chunk)
        chunk_sent_lengths.append([len(_c) for _c in _chunk])
        chunk_idx += 1

    for chunk, chunk_sent_length in tqdm(list(zip(chunks, chunk_sent_lengths)), desc=f"Featurizing {name}"):
        padded_chunk = torch.nn.utils.rnn.pad_sequence(chunk,
                                                       batch_first=True,
                                                       padding_value=0).to(device)
        attention_mask = torch.nn.utils.rnn.pad_sequence(
            [torch.ones(sent_length).long() for sent_length in chunk_sent_length],
            batch_first=True,
            padding_value=0).to(device)
        outs = model(input_ids=padded_chunk,
                     attention_mask=attention_mask,
                     past_key_values=None,
                     output_hidden_states=True,
                     return_dict=True)
        h = []
        for hidden_state, sent_length in zip(outs.hidden_states[-1], chunk_sent_length):
            h.append(hidden_state[sent_length - 1])
        h = torch.stack(h, dim=0)
        feats.append(h.cpu())
    t2 = time.time()
    if verbose:
        print(f'Featurize time: {round(t2-t1, 2)}')
    return torch.cat(feats)
